////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     HCC Module KMP User Implementation
///
///     This file implements the HCC module KMP user. All KMP user related
///     APDUs and ioctls are dealt with in this file.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include "tera_log.h"
#include "tera_pcoip_driver.h"
#include "tera_errors.h"

#include "tera_pcoip_hcc_ioctl.h"
#include "tera_hcc.h"
#include "tera_hcc_internal.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

// ---- Local Variables -----------------------------------------------------

// ---- Local Function Prototypes -------------------------------------------

static INT32 kmp_apdu_send(
    sHCC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    );

static INT32 kmp_ioctl_comm_ctrl_proc(
    sHCC_CBLK    *cblk,
    unsigned long arg
    );

static INT32 kmp_ioctl_pri_context_apdu_get_proc(
    sHCC_CBLK    *cblk,
    unsigned long arg,
    unsigned int  ioctl
    );

static void kmp_pri_context_apdu_proc(
    sHCC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    );

// --------------------------------------------------------------------------
//  kmp_apdu_send
//      Send a KMP APDU
//
//  Description:
//      This function sends a KMP APDU to the blade. The contents of the APDU
//      are copied into an allocated packet and enqueued onto the egress
//      packet queue. The calling function can safely discard the APDU after
//      this function call.
//
static INT32 kmp_apdu_send(
    sHCC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    )
{
    INT32 ret;

    ASSERT(cblk != NULL);

    // Send the APDU
    ret = tera_scp_data_pkt_send(cblk->dev, cblk->scp_handle,
                                 TERA_PCOIP_HCC_APDU_USER_KMP,
                                 &cblk->egress_pkt_q, apdu, apdu_len);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("failed to send APDU (ret=%d)", ret);
        ret = -ERESTARTSYS;
    }

    return ret;
} // kmp_apdu_send()

// --------------------------------------------------------------------------
//  kmp_ioctl_comm_ctrl_proc
//      Process a KMP_COMM_CTRL ioctl
//
//  Description:
//      This function processes a KMP_COMM_CTRL ioctl. This function creates a
//      APDU to instruct the KMP to open/close the HDC channel. The APDU is
//      sent to the blade, which is then forwarded to the KMP on the portal.
//
static INT32 kmp_ioctl_comm_ctrl_proc(
    sHCC_CBLK *cblk,
    unsigned long arg    )
{
    INT32                               ret;
    sTERA_PCOIP_HCC_APDU_KMP_COMM_CTRL *buf;
    sTERA_PCOIP_HCC_APDU_KMP_COMM_CTRL  apdu;

    ASSERT(cblk != NULL);

    buf = (sTERA_PCOIP_HCC_APDU_KMP_COMM_CTRL *)arg;
    if (copy_from_user(&apdu, buf, sizeof(apdu)))
    {
        mTERA_LOG_ERR("copy_from_user failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        // acquire the semaphore preventing simultaneous FW requests
        if(down_interruptible(&cblk->pending_ioctl_lock))
        {
            mTERA_LOG_ERR("down_interruptible failed");
            ret = -ERESTARTSYS;
        }
        else
        {
            // Send the APDU
            ret = kmp_apdu_send(cblk, &apdu, sizeof(apdu));
            if (ret != TERA_SUCCESS)
            {
                mTERA_LOG_ERR("failed to send APDU");
                ret = -ERESTARTSYS;
            }
            up(&cblk->pending_ioctl_lock);
        }
    }

    return ret;
} // kmp_ioctl_comm_ctrl_proc()

// --------------------------------------------------------------------------
//  kmp_ioctl_pri_context_apdu_get_proc
//      Process a KMP_PRI_CTXT_GET ioctl
//
//  Description:
//      This function processes a KMP_PRI_CTXT_GET ioctl. This function creates
//      an APDU to get the negotiated KMP PRI context for the current session
//      from the firmware. The ioctl is put into the pending ioctl queue,
//      where it will be completed when the response comes back.
//
static INT32 kmp_ioctl_pri_context_apdu_get_proc(
    sHCC_CBLK    *cblk,
    unsigned long arg,
    unsigned int  ioctl
    )
{
    INT32                              ret;
    sTERA_PCOIP_HCC_APDU_KMP_PRI_CTXT  apdu;
    sTERA_PCOIP_HCC_APDU_KMP_PRI_CTXT *buf;

    ASSERT(cblk != NULL);

    buf = (sTERA_PCOIP_HCC_APDU_KMP_PRI_CTXT *)arg;
    memset(&apdu, 0, sizeof(apdu));

    // Fill out the APDU
    apdu.apdu_type = TERA_PCOIP_HCC_APDU_KMP_TYPE_PRI_CTXT;

    // acquire the semaphore preventing simultaneous FW requests
    if(down_interruptible(&cblk->pending_ioctl_lock))
    {
        mTERA_LOG_ERR("down_interruptible failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        // Send the APDU
        ret = kmp_apdu_send(cblk, &apdu, sizeof(apdu));
        if (ret != TERA_SUCCESS)
        {
            up(&cblk->pending_ioctl_lock);
            mTERA_LOG_ERR("failed to send APDU");
            ret = -ERESTARTSYS;
        }
        else
        {
            //
            // Set flag equal to the outstanding ioctl, the flag is inspected
            // when the workqueue is inspected.
            //
            cblk->pending_ioctl_flag = ioctl;

            // wait until the response is received on the ingress message queue and
            // processed by the dpc_entry
            if (wait_event_interruptible_timeout(cblk->pending_ioctl_q,
                                                 (cblk->pending_ioctl_flag == 1),
                                                 msecs_to_jiffies(TERA_IOCTL_TIMEOUT_MS)) <= 0)
            {
                mTERA_LOG_ERR("wait_event_interruptible_timeout failed");
                up(&cblk->pending_ioctl_lock);
                cblk->pending_ioctl_flag = 1;
                ret = -ERESTARTSYS;
            }
            else
            {
                // fill in the IOCTL response
                if(copy_to_user(buf, &cblk->ioctl_rsp, sizeof(sTERA_PCOIP_HCC_APDU_KMP_PRI_CTXT)))
                {
                    mTERA_LOG_ERR("copy_to_user failed");
                    ret = -ERESTARTSYS;
                }
                else
                {
                    ret = TERA_SUCCESS;
                }
                up(&cblk->pending_ioctl_lock);
            }
        }
    }
    return ret;
} // kmp_ioctl_pri_context_apdu_get_proc()

// --------------------------------------------------------------------------
//  kmp_pri_context_apdu_proc
//      Process a KMP PRI context APDU
//
//  Description:
//      This function processes a KMP PRI context APDU from the firmware. This
//      function looks for a pending KMP_PRI_CTXT ioctl and copies the PRI
//      context to the output buffer of the ioctl and completes it.
//
static void kmp_pri_context_apdu_proc(
    sHCC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    )
{
    ASSERT(cblk != NULL);

    if ((apdu == NULL) ||
        (apdu_len != sizeof(sTERA_PCOIP_HCC_APDU_KMP_PRI_CTXT)))
    {
        mTERA_LOG_ERR("incorrect APDU length (%u) (expected=%u)",
                      apdu_len, (UINT32)sizeof(sTERA_PCOIP_HCC_APDU_KMP_PRI_CTXT));
    }
    // if the correct IOCTL is not pending
    else if (cblk->pending_ioctl_flag != TERA_PCOIP_HCC_IOCTL_KMP_PRI_CTXT_GET)
    {
        // Ioctl could have been cancelled
        mTERA_LOG_INFO("no pending HCC_IOCTL_KMP_PRI_CTXT_GET");
    }
    else
    {
        down(&cblk->hmi.data_lock);

        // Store the response in the control block
        memcpy(&cblk->ioctl_rsp.data.kmp_pri, apdu,
               sizeof(sTERA_PCOIP_HCC_APDU_KMP_PRI_CTXT));

        up(&cblk->hmi.data_lock);

        // set the flag that will wake up the waiting ioctl thread
        cblk->pending_ioctl_flag = 1;

        // wake up the waiting thread
        wake_up_interruptible(&cblk->pending_ioctl_q);
    }
} // kmp_pri_context_apdu_proc()

// --------------------------------------------------------------------------
//  hcc_kmp_init
//      Initialize the KMP user
//
//  Description:
//      This function initializes the KMP user. This function is responsible
//      for initializing any and all user-specific objects.
//
INT32 hcc_kmp_init(
    sHCC_CBLK *cblk    // HCC control block
    )
{
    INT32 ret;

    mTERA_LOG_DBG("start");

    ASSERT(cblk != NULL);

    // Placeholder for now
    ret = TERA_SUCCESS;

    mTERA_LOG_DBG("end");

    return ret;
} // hcc_kmp_init()

// --------------------------------------------------------------------------
//  hcc_kmp_start
//      Start the HCC KMP user portion
//
//  Description:
//      This function starts the HCC KMP user portion.
//
INT32 hcc_kmp_start(
    sHCC_CBLK *cblk
    )
{
    INT32    ret = TERA_SUCCESS;

    return ret;
} // hcc_kmp_start()

// --------------------------------------------------------------------------
//  hcc_kmp_stop
//      Stop the KMP user
//
//  Description:
//      This function stops the KMP user. This function is responsible
//      for stopping any and all user-specific objects, events, and threads.
//
void hcc_kmp_stop(
    sHCC_CBLK *cblk
    )
{
    INT32                               ret;
    sTERA_PCOIP_HCC_APDU_KMP_COMM_CTRL  apdu;

    ASSERT(cblk != NULL);

    mTERA_LOG_DBG("start");

    // close the HDC channel
    apdu.apdu_type = TERA_PCOIP_HCC_APDU_KMP_TYPE_COMM_CTRL;
    apdu.open_channel = 0;
    // acquire the semaphore preventing simultaneous FW requests
    if(down_interruptible(&cblk->pending_ioctl_lock))
    {
        mTERA_LOG_ERR("down_interruptible failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        ret = kmp_apdu_send(cblk, &apdu, sizeof(apdu));
        if (ret != TERA_SUCCESS)
        {
            up(&cblk->pending_ioctl_lock);
            mTERA_LOG_ERR("failed to send APDU");
        }
        else
        {
            up(&cblk->pending_ioctl_lock);
            mTERA_LOG_DBG("HDC channel closed");
        }
    }
// TBD - I tried waiting a little to ensure the channel closed but it didn't work...
//     ssleep(5);
} // hcc_kmp_stop()

// --------------------------------------------------------------------------
//  hcc_kmp_ioctl_proc
//      Process an HCC KMP user ioctl
//
//  Description:
//      This function processes a HCC KMP user ioctl. If the ioctl is
//      processed, this function returns TERA_SUCCESS. If the ioctl is unknown,
//      this function returns TERA_ERR_NOT_SUPPORTED.
//
INT32 hcc_kmp_ioctl_proc(
    sHCC_CBLK      *cblk,
    unsigned long   arg,
    unsigned int    ioctl
    )
{
    INT32 ret;

    ASSERT(cblk != NULL);

    switch (ioctl)
    {
        case TERA_PCOIP_HCC_IOCTL_KMP_HDC_COMM_CTRL:
        {
            mTERA_LOG_DBG("HCC_IOCTL_KMP_HDC_COMM_CTRL");

            ret = kmp_ioctl_comm_ctrl_proc(cblk, arg);
            break;
        }
        case TERA_PCOIP_HCC_IOCTL_KMP_PRI_CTXT_GET:
        {
            mTERA_LOG_DBG("HCC_IOCTL_KMP_PRI_CTXT_GET");

            ret = kmp_ioctl_pri_context_apdu_get_proc(cblk, arg, ioctl);
            break;
        }
        default:
        {
            mTERA_LOG_DBG("unsupported ioctl=0x%08x", ioctl);
            ret = TERA_ERR_NOT_SUPPORTED;
            break;
        }
    }
    return ret;
} // hcc_kmp_ioctl_proc()

// --------------------------------------------------------------------------
//  hcc_kmp_apdu_proc
//      Process a KMP APDU
//
//  Description:
//      This function processes a KMP APDU from the ingress descriptor queue.
//      The APDU is freed by the calling function after this function returns,
//      so it should not be referenced for future use.
//
void hcc_kmp_apdu_proc(
    sHCC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    )
{
    eTERA_PCOIP_HCC_APDU_KMP_TYPE *apdu_type = apdu;

    ASSERT(cblk != NULL);
    ASSERT(apdu != NULL);

    if (apdu_len < sizeof(eTERA_PCOIP_HCC_APDU_KMP_TYPE))
    {
        mTERA_LOG_ERR("invalid APDU length");
    }
    else if (cblk->initialized)
    {
        switch (*apdu_type)
        {
            case TERA_PCOIP_HCC_APDU_KMP_TYPE_PRI_CTXT:
            {
                mTERA_LOG_DBG("KMP_TYPE_PRI_CTXT");

                kmp_pri_context_apdu_proc(cblk, apdu, apdu_len);
                break;
            }
            default:
            {
                mTERA_LOG_ERR("unsupported KMP APDU type=%u", *apdu_type);
                break;
            }
        }
    }
} // hcc_kmp_apdu_proc()

