////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     Host Data Channel Module Implementation
///
///     This file implements the Host Data Channel (HDC) module. This
///     module implements a messaging system for communication between the
///     firmware and host OS. User applications on the host OS can use ioctl
///     calls to send packets to the HDC module. The packets are then sent to
///     the firmware via the GFUN module.
///
///     This module depends on the GFUN, message, memory pool, and packet
///     modules. Those modules MUST be initialized before calling this
///     module's initialization function.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include "tera_log.h"
#include "tera_pcoip_driver.h"
#include "tera_errors.h"
#include "tera_pcoip_hdc_ioctl.h"
#include "tera_hdc.h"
#include "tera_hdc_internal.h"
#include "tera_gfun.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

// ---- Local Variables -----------------------------------------------------

// This workqueue is used to schedule a kernel thread that periodically polls
// the PCoIP HDC device.
static struct workqueue_struct *hdc_wq;

// The following pointer is initialized when the driver starts. The pointer is
// referenced by the dpc_entry function.
//
// Note: It would be better to pass the pointer into the function but
// workqueues on some Linux distributions do not support passing in a
// parameter. This feature was removed in version 2.6.20 of the kernel.
//
sTERA_DEV_CBLK *hdc_dev_cblk = NULL;

// This variable tracks the maximum number of KMP messages received from
// the blade and processed by the HDC thread. This can be used to determine
// whether messages are not being processed quickly, which can lead to
// delay in the performance of the desktop cursor.
//
UINT32 max_rx_kmp_msgs_processed = 0;

// ---- Local Function Prototypes -------------------------------------------

static void scp_chan_open_cback(
    UINT8 chan_id,
    void *user
    );

static INT hdc_egress_desc_q_proc(
    sHDC_CBLK *cblk
    );

static void hdc_ingress_desc_q_proc(
    sHDC_CBLK      *cblk,
    sTERA_MEM_BUF  *buf,
    UINT32          buf_len
    );

static void hdc_thread_entry(
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20))
    struct work_struct *unused
#else
    void *unused
#endif
    );

// --------------------------------------------------------------------------
//  scp_chan_open_cback
//      SCP channel open callback
//
//  Description:
//      This function is a callback for the SCP channel open event.
//
static void scp_chan_open_cback(
    UINT8 chan_id,
    void *user
    )
{
    sHDC_CBLK *cblk = user;

    ASSERT(user != NULL);

    switch (chan_id)
    {
        case TERA_PCOIP_HDC_APDU_USER_KMP:
        {
        	// reset the counter at the start of a session
            max_rx_kmp_msgs_processed = 0;

            mTERA_LOG_INFO("KMP SCP channel open");

            hdc_kmp_scp_chan_open_proc(cblk);
            break;
        }
        default:
        {
            //
            // We should never be here because we only tell the SCP module to
            // open SCP channels that we know of
            //
            mTERA_LOG_ERR("unsupported SCP channel ID=%u", chan_id);
            break;
        }
    }
} // scp_chan_open_cback()

// --------------------------------------------------------------------------
//  hdc_egress_desc_q_proc
//      Send a packet to the egress descriptor queue
//
//  Description:
//      This function sends an outstanding packet (if exists) to the egress
//      descriptor queue. The packet is pulled from the egress packet
//      queue.
//
static INT hdc_egress_desc_q_proc(
    sHDC_CBLK *cblk
    )
{
    INT        ret = 0;
    INT32      err;
    sTERA_PKT *pkt;

    if ((cblk != NULL) && cblk->initialized)
    {
        // Grab a packet from the head of the queue
        pkt = tera_pkt_q_dequeue(&cblk->egress_pkt_q);
        if (pkt != NULL)
        {
#ifdef GFUN_DBG
            sTERA_MEM_BUF *buf = pkt->buf;
            UINT32 *data_ptr = buf->virt_addr;
            mTERA_LOG_DBG("buf data=%08X %08X %08X %08X %08X %08X %08X %08X",
                          data_ptr[0], data_ptr[1], data_ptr[2], data_ptr[3],
                          data_ptr[4], data_ptr[5], data_ptr[6], data_ptr[7]);
#endif

            // Send the packet off to the firmware
            err = tera_gfun_tx(cblk->dev, pkt->buf, pkt->buf_len, NULL);
            if (err != TERA_SUCCESS)
            {
                mTERA_LOG_ERR("failed to send packet over GFUN (ret=%d)", ret);

                if (err == TERA_ERR_FULL)
                {
                    //
                    // If the send failed because the egress descriptor queues
                    // are full, then we should add the packet back to onto the
                    // egress packet queue to try to send it again later.
                    //
                    tera_pkt_q_head_enqueue(&cblk->egress_pkt_q, pkt);
                }
                else
                {
                    //
                    // We're here because the send failed and its not because
                    // the egress descriptor queues are full. This must mean
                    // something bad has happened or we're in the process of
                    // being shutdown. Thus, drop the packet on the floor.
                    //
                    tera_pkt_free(cblk->dev, pkt);
                }
            }
            else
            {
#ifdef GFUN_DBG
                mTERA_LOG_DBG("sent packet to GFUN with len=%d", pkt->buf_len);
#endif

                // The GFUN owns the buffer now, so detach it from the packet
                pkt->buf = NULL;

                // Free the packet
                tera_pkt_free(cblk->dev, pkt);

                ret = 1;
            }
        } // end if
    } // end if

    return ret;
} // hdc_egress_desc_q_proc()

// --------------------------------------------------------------------------
//  hdc_ingress_desc_q_proc
//      Process a packet that came from the ingress descriptor queue
//
//  Description:
//      This function processes a packet on the ingress descriptor queue. The
//      packet is MUXed and sent to the appropriate HDC user to process.
//
static void hdc_ingress_desc_q_proc(
    sHDC_CBLK     *cblk,       // Control block
    sTERA_MEM_BUF *buf,        // Packet buffer
    UINT32         buf_len     // Packet buffer length
    )
{
    UINT8  chan_id;
    UINT8 *apdu = NULL;
    UINT32 apdu_len;

    if (buf == NULL)
    {
        mTERA_LOG_ERR("received 0-byte packet");
    }
    else if ((cblk != NULL) && cblk->initialized)
    {
        tera_scp_pkt_proc(cblk->dev, cblk->scp_handle, buf, buf_len, &chan_id,
                          &apdu, &apdu_len, &cblk->egress_pkt_q);

        // Continue processing if we have data (non-NULL APDU pointer)
        if (apdu != NULL)
        {
            // The SCP channel ID is the same as the HCC user ID
            switch (chan_id)
            {
                case TERA_PCOIP_HDC_APDU_USER_KMP:
                {
#ifdef GFUN_DBG
                    mTERA_LOG_DBG("HDC_APDU_USER_KMP");
#endif

                    hdc_kmp_apdu_proc(cblk, apdu, apdu_len);
                    break;
                }
                default:
                {
                    mTERA_LOG_ERR("invalid HDC user (chan_id=%u)", chan_id);
                    break;
                }
            }
        }
    } // end if

    // Free the packet buffer
    tera_mem_buf_free(cblk->dev, buf);
} // hdc_ingress_desc_q_proc()

// --------------------------------------------------------------------------
//  hdc_thread_entry
//      Entry point to the deferred procedure call (DPC) to process packets
//
//  Description:
//      The DPC will process in the following priority:
//          1. ingress descriptor queue packets
//          2. egress packet queue packets
//
static void hdc_thread_entry(
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20))
    struct work_struct *unused
#else
    void *unused
#endif
    )
{
    sHDC_CBLK     *cblk;
    INT32          err;
    INT            processed;
    INT            kmp_processed;
    sTERA_MEM_BUF *buf;
    UINT32         buf_len;
    UINT32         rx_kmp_msgs = 0;

    if(hdc_dev_cblk != NULL)
    {
        cblk = &hdc_dev_cblk->hdc_cblk;

        do
        {
            if (!cblk->started)
            {
                //
                // We could still be running when the module is being stopped
                // because the DPC runs on a different thread than the rest of the
                // driver. In that case, we need to refrain from doing any
                // processing.
                //
                break;
            }

            processed = 0;

            // Check if there is room on the egress descriptor queue
            err = tera_gfun_tx_ready(cblk->dev);
            if (err == TERA_SUCCESS)
            {
                // Send a packet, if any, from the egress packet queue
                processed |= hdc_egress_desc_q_proc(cblk);
            }
            else if (err == TERA_ERR_FULL)
            {
                mTERA_LOG_DBG("egress descriptor queue full");
            }
            else
            {
                // Stop processing on all other errors
                break;
            }

            // Recieve a packet, if available, from the ingress descriptor queue
            if (tera_gfun_rx(cblk->dev, &buf, &buf_len, NULL) == TERA_SUCCESS)
            {
                // Process the packet
                hdc_ingress_desc_q_proc(cblk, buf, buf_len);

                processed |= 1;
            }

            // Carry out any KMP processing
            kmp_processed = hdc_kmp_thread_proc(cblk);
            if (kmp_processed)
            {
                processed |= kmp_processed;
                rx_kmp_msgs++;
            }

#ifdef GFUN_DBG
            // if any messages were processed dump the GFUN registers
            if (processed)
            {
                tera_gfun_reg_dump(cblk->dev);
            }
#endif
        } while (processed);

        // re-schedule the dpc
        queue_delayed_work(hdc_wq, &cblk->work, msecs_to_jiffies(IRQ_TIMER_PERIOD_MS));

        if (rx_kmp_msgs > max_rx_kmp_msgs_processed)
        {
            max_rx_kmp_msgs_processed = rx_kmp_msgs;
            mTERA_LOG_DBG("max_rx_kmp_msgs_processed = %d",
                          max_rx_kmp_msgs_processed);
        }
    }
    else
    {
        mTERA_LOG_CRIT("hdc_thread_entry started before initializing hdc_dev_cblk");
    }
} // hdc_thread_entry()

// --------------------------------------------------------------------------
//  tera_hdc_init_phase1
///     Phase 1 initialization of the Host Data Channel (HDC) module
//
//  Description:
///     This function initializes portions of the HDC module. The device
///     control block has both an HCC and HDC control block. Initially the
///     driver does not know if it is communicating with an HCC or HDC. The
///     phase 1 initialization functions are invoked for both the HCC and HDC
///     control blocks within the device control block. After phase 1
///     initialization completes the code can read a register to determine
///     whether it is communicating with an HCC or HDC. At that point the code
///     should call the appropriate phase 2 initialization function.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_FAILURE
///
INT32 tera_hdc_init_phase1(
    struct  pci_dev *dev    ///< Device object
    )
{
    INT32                 ret = TERA_ERR_FAILURE;
    sTERA_DEV_CBLK       *dev_cblk = pci_get_drvdata(dev);
    sHDC_CBLK            *cblk=&dev_cblk->hdc_cblk;

    mTERA_LOG_DBG("start");

    // Initialize control block variables
    cblk->dev = dev;
    cblk->pending_ioctl_flag = 0;

    // Create a pending ioctl queue
    init_waitqueue_head(&cblk->pending_ioctl_q);

    // Create a spinlock that ensures only one request is sent to the FW
    init_MUTEX(&cblk->pending_ioctl_lock);

    // Create a spinlock that protects the HDC data structures
    init_MUTEX(&cblk->data_lock);

    // Initialize the egress packet queue
    ret = tera_pkt_q_init(&cblk->egress_pkt_q);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("failed to initialize egress packet queue (ret=%d)",
                      ret);
    }
    else
    {
        // Create a SCP instance for the HDC
        ret = tera_scp_create(dev, (1 << TERA_PCOIP_HDC_APDU_USER_KMP),
                              &cblk->scp_handle, scp_chan_open_cback, cblk);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("failed to create SCP instance (ret=%d)", ret);
        }
    }

    mTERA_LOG_DBG("end (ret=%d)", ret);

    return ret;
} // tera_hdc_init_phase1()

// --------------------------------------------------------------------------
//  tera_hdc_init_phase2
///     Phase 2 initialization of the Host Data Channel (HDC) module
//
//  Description:
///     See the comments that describe the tera_hdc_init_phase1() function.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_FAILURE
///
INT32 tera_hdc_init_phase2(
    struct  pci_dev *dev    ///< Device object
    )
{
    INT32                 ret = TERA_ERR_FAILURE;
    sTERA_DEV_CBLK       *dev_cblk = pci_get_drvdata(dev);
    sHDC_CBLK            *cblk=&dev_cblk->hdc_cblk;

    mTERA_LOG_DBG("start");

    // Initialize control block variables
    hdc_dev_cblk = dev_cblk;

    // Call the KMP-user initialization routine
    ret = hdc_kmp_init(cblk);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("failed to initialize the KMP user (ret=%d)", ret);
    }
    else
    {
        // Create a workqueue for the hdc_thread_entry function that polls the device
        hdc_wq = create_singlethread_workqueue("hdc_wq");
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20))
        INIT_DELAYED_WORK(&cblk->work, hdc_thread_entry);
#else
        INIT_WORK(&cblk->work, hdc_thread_entry, NULL);
#endif

        // All systems green
        cblk->initialized = 1;
    }

    mTERA_LOG_DBG("end (ret=%d)", ret);

    return ret;
} // tera_hdc_init_phase2()

// --------------------------------------------------------------------------
//  tera_hdc_cleanup
///     HDC module clean up routine
//
//  Description:
///     This function performs the HDC module clean up routine. The system
///     timer resolution is reset back to its default value, and the module is
///     no longer marked as being initialized.
///
/// \return
///     None
///
void tera_hdc_cleanup(
    struct  pci_dev *dev    ///< Device object
    )
{
    sTERA_DEV_CBLK *dev_cblk = pci_get_drvdata(dev);
    sHDC_CBLK      *cblk=&dev_cblk->hdc_cblk;

    mTERA_LOG_DBG("start");

    if (cblk->initialized)
    {
        cblk->initialized = 0;

        // Call the KMP user cleanup routine
        hdc_kmp_cleanup(cblk);
    }

    mTERA_LOG_DBG("end");
} // tera_hdc_cleanup()

// --------------------------------------------------------------------------
//  tera_hdc_start
///      Start the HDC module
//
//  Description:
///     This function starts the HDC module. All local variable values
///     will be reset. If there were any pending feature notify ioctls, then
///     this function will complete them with a cleared feature set.
///
/// \return
///     TERA_SUCCESS
///     TERA_ERR_INVALID_STATE
///     TERA_ERR_FAILURE
///
INT32 tera_hdc_start(
    struct  pci_dev *dev    ///< Device object
    )
{
    INT32      ret = TERA_ERR_FAILURE;
    sTERA_DEV_CBLK *dev_cblk = pci_get_drvdata(dev);
    sHDC_CBLK      *cblk=&dev_cblk->hdc_cblk;

    mTERA_LOG_DBG("start");

    if (!cblk->initialized)
    {
        mTERA_LOG_ERR("HDC module not initialized");

        ret = TERA_ERR_INVALID_STATE;
    }
    else
    {
        if (cblk->started)
        {
            mTERA_LOG_DBG("HDC module already started");

            ret = TERA_SUCCESS;
        }
        else
        {
#if 0  // TBD - 1.4.5 code
            //
            // Reset the timestamps used for local cursor and keyboard HID read
            // report injection
            //
            cblk->lcur_next_tstamp = 0;
            cblk->lkbd_next_tstamp = 0;
#endif

            // Start the HDC KMP user
            ret = hdc_kmp_start(cblk);
            if (ret != TERA_SUCCESS)
            {
                mTERA_LOG_ERR("failed to start HDC KMP user (ret=%d)", ret);
            }
            else
            {
                // Schedule the DPC to run to send messages to Tera host
                queue_delayed_work(hdc_wq, &cblk->work, msecs_to_jiffies(IRQ_TIMER_PERIOD_MS));
                mTERA_LOG_INFO("HDC thread sched period = %lu jiffies, ticks/sec = %d, ticks/10 ms = %d",
                               msecs_to_jiffies(IRQ_TIMER_PERIOD_MS), HZ, HZ/100);

                // Ready to go!
                cblk->started = 1;
            }
        }
    }

    mTERA_LOG_DBG("end");

    return ret;
} // tera_hdc_start()

// --------------------------------------------------------------------------
//  tera_hdc_stop
///     Stop the HDC module
//
//  Description:
///     This function stops the HDC module. All packets from the egress and
///     ingress descriptor queues, and the local cursor queue are freed.
///
/// \return
///     None
///
void tera_hdc_stop(
    struct  pci_dev *dev    ///< Device object
    )
{
    sTERA_DEV_CBLK *dev_cblk = pci_get_drvdata(dev);
    sHDC_CBLK      *cblk=&dev_cblk->hdc_cblk;

    mTERA_LOG_DBG("start");

    if (cblk->initialized && cblk->started)
    {
        // Stop the KMP user
        hdc_kmp_stop(cblk);

        //
        // Clear the started flag first- even if there is a error while
        // stopping the module, we'll still end up in the stopped state anyway.
        //
        cblk->started = 0;

        // Stop the hdc_thread_entry thread
        cancel_delayed_work(&cblk->work);
        flush_workqueue(hdc_wq);
        destroy_workqueue(hdc_wq);

        // Delete all packets from the egress packet queue
        tera_pkt_q_purge(dev, &cblk->egress_pkt_q);

        //
        // Reset the SCP - we do this AFTER we stop the users because the
        // users may still need to send packets to the firmware
        //
        tera_scp_reset(cblk->scp_handle);
    } // end if

    mTERA_LOG_DBG("end");
} // tera_hdc_stop()

// --------------------------------------------------------------------------
//  tera_hdc_ioctl_proc
///     Process an ioctl request for the HDC device interface
//
//  Description:
///     This function processes an ioctl request for the HDC device interface.
///     If the ioctl is unknown, then it is completed with STATUS_NOT_SUPPORTED.
///
/// \return
///     None
///
INT32 tera_hdc_ioctl_proc(
    struct  pci_dev *dev,
    unsigned int  ioctl,
    unsigned long arg
    )
{
    INT32           ret;
    sTERA_DEV_CBLK *dev_cblk = pci_get_drvdata(dev);
    sHDC_CBLK      *cblk = &dev_cblk->hdc_cblk;

    if ((cblk == NULL) || !cblk->initialized)
    {
        mTERA_LOG_ERR("not initialized");
        ret = TERA_ERR_INVALID_ARG;
    }
    else if (!cblk->started)
    {
        mTERA_LOG_ERR("not started");
        ret = TERA_ERR_INVALID_STATE;
    }
    else
    {
        // Switch based on the user ID encoded into the ioctl
        switch (mTERA_PCOIP_DEV_IOCTL_USER_ID_GET(ioctl))
        {
            case TERA_PCOIP_HDC_APDU_USER_KMP:
            {
                ret = hdc_kmp_ioctl_proc(cblk, arg, ioctl);
                break;
            }
            default:
            {
                mTERA_LOG_DBG("unsupported ioctl (0x%08x) (user=%u)", ioctl,
                              mTERA_PCOIP_DEV_IOCTL_USER_ID_GET(ioctl));

                ret = TERA_ERR_NOT_SUPPORTED;
                break;
            }
        }
    }

    return ret;
} // tera_hdc_ioctl_proc()


