////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     Monitor Class Implementation
///
///     This file implements the monitor class. This class can detect all
///     active display monitors and their following properties:
///         - position
///         - resolution
///         - name of adapter that the monitor is connected to
///         - EDID information:
///             - vendor ID
///             - product ID
///             - serial number
///             - monitor name
///
///     This class can detect if the monitor configuration has changed and
///     it can push client topology settings into the host OS.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "tera_types.h"
#include "tera_log.h"
#include "tera_errors.h"
#include "tera_mem.h"
#include "tera_disp.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

// This is the GUID of the Windows monitor device setup class
//DEFINE_GUID (GUID_CLASS_MONITOR, 0x4d36e96e, 0xe325, 0x11ce, 0xbf, 0xc1, 0x08, 0x00, 0x2b, 0xe1, 0x03, 0x18);

//
// Monitor device interface
// {E6F07B5F-EE97-4a90-B076-33F57BF4EAA7}
//DEFINE_GUID(GUID_DEVINTERFACE_MONITOR, 0xe6f07b5f, 0xee97, 0x4a90, 0xb0, 0x76, 0x33, 0xf5, 0x7b, 0xf4, 0xea, 0xa7);

// EDID block defines
#define EDID_HDR                     0x00FFFFFFFFFFFF00
#define EDID_DESC_BLOCK1_OFFSET      0x36
#define EDID_DESC_BLOCK2_OFFSET      0x48
#define EDID_DESC_BLOCK3_OFFSET      0x5A
#define EDID_DESC_BLOCK4_OFFSET      0x6C
#define EDID_DESC_BLOCK_DATA_OFFSET  0x5
#define EDID_DESC_BLOCK_HDR_SERIAL   0xFF000000
#define EDID_DESC_BLOCK_HDR_NAME     0xFC000000
#define EDID_VID_BYTE1_OFFSET        0x8
#define EDID_VID_BYTE2_OFFSET        0x9
#define EDID_PID_BYTE1_OFFSET        0xA
#define EDID_PID_BYTE2_OFFSET        0xB
#define EDID_SERIAL_OFFSET           0xC

//
// This constant is used to indicate the software could not match the client
// supplied monitor topology with a monitor attached to a display adapter.
//
#define NO_MAPPING_FOUND            (-1)

//
// This constant indicates the software found an active display adapter without
// an attached monitor. In this situation the adapter should be turned off.
//
#define ADAPTER_WITHOUT_MONITOR     (-2)

//
// Macro to compare two different refresh rates
//
// The macro returns a non-zero value if the two rates are equal to each other
// or within 1 of each other. This check is included because there can be some
// round off error when applying a specific refresh rate to the display
// settings.
//
// This check is included to deal with a problem observed on one system with
// a 1680x1050 display. While using the Nvidia Control panel I changed the
// display refresh rate from 59 to 60. When the new rate (60) was applied
// the Control Panel indicated the new setting was applied without an error
// but the refresh rate reported by the Control Panel remained equal to 59.
//
#define mTERA_REFRESH_RATES_SAME(r1, r2) \
    (((r1) >= ((r2)-1)) && ((r1) <= ((r2)+1)))


// ---- Local Variables -----------------------------------------------------

// ---- Local Function Prototypes -------------------------------------------

// --------------------------------------------------------------------------
//  cTERA_DISP::cTERA_DISP
///     Constructor
//
//  Description:
///     This function is the class contructor and initializes the member data
///     variables.
///
/// \return
///     None
///
cTERA_DISP::cTERA_DISP(void)
{
    mTERA_LOG_DBG("start");

    // Initialize member variables
    m_disp_cnt = 0;

    mTERA_LOG_DBG("end");
} // cTERA_DISP::cTERA_DISP()

// --------------------------------------------------------------------------
//  cTERA_DISP::~cTERA_DISP
///     Destructor
//
//  Description:
///     This function performs clean up of this class, such as freeing the
///     memory used for the host monitor list.
///
/// \return
///     None
///
cTERA_DISP::~cTERA_DISP(void)
{
    mTERA_LOG_DBG("start-end");
} // cTERA_DISP::~cTERA_DISP()

// --------------------------------------------------------------------------
//  cTERA_DISP::disp_topology_get
///     Get the display topology
//
//  Description:
///     This function gets the display topology. Only active, non-mirrored
///     displays are returned. The maximum number of displays returned by this
///     function is TERA_DISP_MAX_DISP. If there are more than that number of
///     active displays, only the first TERA_DISP_MAX_DISP displays are
///     returned.
///
///     If the agent is talking to a client that does not provide EDID
///     information (pre-3.2.x) and multiple displays are active the displays
///     are sorted in order from L->R, T->B.
///
///     If the agent is talking to a 3.2.x or higher client the displays are
///     sorted by client DVI port starting with DVI-0.
///
INT32 cTERA_DISP::disp_topology_get(
    sTERA_DISP_TOPO    *disp_topo,
    sTERA_DISP_EDID    *disp_edid,
    sTERA_DISP_EDID    *client_edid
    )
{
#if 1 // TBD - need to implement
    return TERA_SUCCESS;
#else
    INT32               ret = TERA_ERR_FAILURE;
    UINT32              num_disp = 0;
    UINT32              disp_cnt = 0;
    DISPLAY_DEVICE      adapter;
    DISPLAY_DEVICE      mon;
    sTERA_DISP          temp_disp;
    sTERA_DISP_EDID_ID  temp_edid;
    sTERA_DISP_EDID_ID  mon_edid_id;
    DEVMODE             mode;
    UINT32              i = 0;
    UINT32              j;
    INT                 disp_settings_found;
    INT                 monitor_found;

    mTERA_LOG_DBG("start");

    if ((disp_topo == NULL) || (disp_edid == NULL) || (client_edid == NULL))
    {
        mTERA_LOG_ERR("invalid argument");
        ret = TERA_ERR_INVALID_ARG;
    }
    else
    {
        for (j=0; j<TERA_DISP_MAX_DISP; j++)
        {
            mTERA_LOG_DBG(" client EDID%u serial: %s",
                          j,
                          client_edid->edid[j].serial);
        }

        mTERA_MEM_ZERO(disp_topo, sizeof(sTERA_DISP_TOPO));
        mTERA_MEM_ZERO(disp_edid, sizeof(sTERA_DISP_EDID));

        // Get the number of active displays
        num_disp = GetSystemMetrics(SM_CMONITORS);

        mTERA_LOG_DBG("found %u display(s)", num_disp);

		mTERA_MEM_ZERO(&adapter, sizeof(adapter));
        adapter.cb = sizeof(adapter);

        // Go through all display adapters
        while (EnumDisplayDevices(NULL, i++, &adapter, 0))
        {
            dump_display_device(&adapter);

            if (!(adapter.StateFlags & DISPLAY_DEVICE_MIRRORING_DRIVER) &&
				 (adapter.StateFlags & DISPLAY_DEVICE_ATTACHED_TO_DESKTOP))
            {
				mTERA_LOG_DBG(" attached monitor!!!");

                disp_settings_found = 1;
                mTERA_MEM_ZERO(&mode, sizeof(DEVMODE));
                mode.dmSize = sizeof(DEVMODE);

                // Get the display setting from cache...
                if (!EnumDisplaySettings(adapter.DeviceName,
                                         ENUM_CURRENT_SETTINGS,
                                         &mode))
                {
                    mTERA_LOG_DBG("Unable to get display settings from the cache (ret=%u), trying registry",
                                  GetLastError());

                    // Try to get the display settings from the registry
                    if (!EnumDisplaySettings(adapter.DeviceName,
                                             ENUM_REGISTRY_SETTINGS,
                                             &mode))
                    {
                        mTERA_LOG_ERR("failed to get display settings from registry %u",
                                      GetLastError());
                        disp_settings_found = 0;
                    }
                }

                // Look for the monitor attached to the adapter
                mTERA_MEM_ZERO(&mon, sizeof(DISPLAY_DEVICE));
                mon.cb = sizeof(DISPLAY_DEVICE);
                j = 0;
                monitor_found = 0;
                while (EnumDisplayDevices(adapter.DeviceName,
                                          j++, &mon, EDD_GET_DEVICE_INTERFACE_NAME))
                {
                    mTERA_LOG_DBG(" mon.StateFlags = 0x%08X", mon.StateFlags);

                    // If a monitor is attached parse its EDID
                    if (is_monitor_attached(&mon))
                    {
                        dump_display_device(&mon);

                        // Parse the EDID identification data
                        if (edid_parse(mon.DeviceKey, &mon_edid_id))
                        {
                            monitor_found = 1;
                        }
                        break;
                    }
                    else
                    {
                        mTERA_LOG_DBG("  no attached monitor detected!!!");
                        if (edid_parse(mon.DeviceKey, &mon_edid_id))
                        {
                            mTERA_LOG_DBG("   edid_parse returned TRUE");
                        }
                        else
                        if (edid_parse(mon.DeviceKey, &mon_edid_id))
                        {
                            mTERA_LOG_DBG("   edid_parse returned FALSE");
                        }
                    }
                }
                if (!monitor_found)
                {
                    mTERA_LOG_DBG("  unable to find the attached monitor!!!");
                }
                if (disp_settings_found && monitor_found)
                {
                    // Copy the display into the user buffer
                    disp_topo->disp[disp_cnt].x_origin = mode.dmPosition.x;
                    disp_topo->disp[disp_cnt].y_origin = mode.dmPosition.y;

                    if (mode.dmFields & DM_DISPLAYORIENTATION)
                    {
                        disp_topo->disp[disp_cnt].rotation =
                            (mode.dmDisplayOrientation == DMDO_90) ?
                            TERA_DISP_ROTATION_90 :
                            (mode.dmDisplayOrientation == DMDO_180) ?
                            TERA_DISP_ROTATION_180 :
                            (mode.dmDisplayOrientation == DMDO_270) ?
                            TERA_DISP_ROTATION_270 : TERA_DISP_ROTATION_0;
                    }
                    // assume rotation = 0 if the dmFields rotation bit is clear
                    else
                    {
                        disp_topo->disp[disp_cnt].rotation = TERA_DISP_ROTATION_0;

                        mTERA_LOG_DBG("DM_DISPLAYORIENTATION clear in mode.dmFields, assuming disp%u rotation=0",
                                      disp_cnt);
                    }

                    //
                    // The client device expects the width and height to be the
                    // pre-rotation values but Windows reports post rotation
                    // values so the code swaps them if rot != 0 or 180
                    //
                    if ((disp_topo->disp[disp_cnt].rotation == TERA_DISP_ROTATION_0) ||
                        (disp_topo->disp[disp_cnt].rotation == TERA_DISP_ROTATION_180))
                    {
                        disp_topo->disp[disp_cnt].width = (UINT16)mode.dmPelsWidth;
                        disp_topo->disp[disp_cnt].height = (UINT16)mode.dmPelsHeight;
                    }
                    else
                    {
                        disp_topo->disp[disp_cnt].width = (UINT16)mode.dmPelsHeight;
                        disp_topo->disp[disp_cnt].height = (UINT16)mode.dmPelsWidth;
                    }

                    if (mode.dmFields & DM_DISPLAYFREQUENCY)
                    {
                        disp_topo->disp[disp_cnt].refresh_rate = (UINT8)mode.dmDisplayFrequency;
                    }
                    // assume refresh rate = 0 if the dmFields frequency bit is clear
                    else
                    {
                        disp_topo->disp[disp_cnt].refresh_rate = 0;

                        mTERA_LOG_DBG("DM_DISPLAYFREQUENCY clear in mode.dmFields, assuming ref=0",
                                      disp_cnt);
                    }

                    // save the monitor EDID information
                    strcpy_s(disp_edid->edid[disp_cnt].vid,
                             TERA_MON_EDID_VID_LEN,
                             mon_edid_id.vid);
                    strcpy_s(disp_edid->edid[disp_cnt].pid,
                             TERA_MON_EDID_PID_LEN,
                             mon_edid_id.pid);
                    strcpy_s(disp_edid->edid[disp_cnt].serial,
                             TERA_MON_EDID_SERIAL_LEN,
                             mon_edid_id.serial);
                    strcpy_s(disp_edid->edid[disp_cnt].name,
                             TERA_MON_EDID_NAME_LEN,
                             mon_edid_id.name);

                    mTERA_LOG_DBG("disp%u: %d x %d @ (%d, %d) rot=%u", disp_cnt,
                                  disp_topo->disp[disp_cnt].width,
                                  disp_topo->disp[disp_cnt].height,
                                  disp_topo->disp[disp_cnt].x_origin,
                                  disp_topo->disp[disp_cnt].y_origin,
                                  disp_topo->disp[disp_cnt].rotation);
                    mTERA_LOG_DBG("        rotation=%u, freq=%d, dmFields=0x%08X",
                                  disp_topo->disp[disp_cnt].rotation == TERA_DISP_ROTATION_90 ? 90 :
                                  disp_topo->disp[disp_cnt].rotation == TERA_DISP_ROTATION_180 ? 180 :
                                  disp_topo->disp[disp_cnt].rotation == TERA_DISP_ROTATION_270 ? 270 : 0,
                                  disp_topo->disp[disp_cnt].refresh_rate, mode.dmFields);
                    mTERA_LOG_DBG("        name=%s, serial=%s, vid=%s, pid=%s",
                                  disp_edid->edid[disp_cnt].name,
                                  disp_edid->edid[disp_cnt].serial,
                                  disp_edid->edid[disp_cnt].vid,
                                  disp_edid->edid[disp_cnt].pid);

                    // Increment the display count
                    disp_cnt++;
                }
                else
                {
                    // Must have hit case where monitor was disabled just now
                    mTERA_LOG_DBG("found active adapter, but no attached monitor");
                }
            }
            else
            {
                // Ignore display adapters not used in the desktop
                mTERA_LOG_DBG("  ignoring unused display (%s)",
                              adapter.DeviceString);
            }

            mTERA_MEM_ZERO(&adapter, sizeof(adapter));
            adapter.cb = sizeof(adapter);
        } // end while

        // Fill out the last piece of information
        disp_topo->num_disp = disp_cnt;

        //
        // If two displays are active and the client does not provide EDID
        // data we're talking to a soft client or a zero client running old
        // firmware (pre 3.2.0). In this situation the displays must be
        // arranged so that the left/top most display is connected to the
        // lowest numbered DVI port. The following code swaps the displays
        // if necessary.
        //
        if (client_edid->num_disp == 0)
        {
            mTERA_LOG_DBG("Ordering displays from Left->Right, Top->Bottom");

            if (num_disp == 2)
            {
                //
                // Sort the displays in the user buffer. The displays need
                // to be put into the buffer in a sorted order from L->R,
                // T->B.
                //
                INT32      x1 = disp_topo->disp[0].x_origin;
                INT32      x2 = disp_topo->disp[1].x_origin;
                INT32      y1 = disp_topo->disp[0].y_origin;
                INT32      y2 = disp_topo->disp[1].y_origin;
                INT32      max_x;
                INT32      max_y;
                max_x = (disp_topo->disp[0].width > disp_topo->disp[1].width) ?
                         disp_topo->disp[0].width : disp_topo->disp[1].width;
                max_y = (disp_topo->disp[0].height > disp_topo->disp[1].height) ?
                         disp_topo->disp[0].height : disp_topo->disp[1].height;

                //
                // The first line handles the case where the displays are in the
                // same horizontal plane and the second line handles displays in
                // the same vertical plane (stack on top of each other)
                //
                if(((y1 <= max_y) && (y2 <= max_y) && (x1 > x2)) ||
                   ((x1 <= max_x) && (x2 <= max_x) && (y1 > y2)))
                {
                    mTERA_LOG_DBG("swapping displays - new order: disp0 (%d,%d), disp1 (%d,%d)",
                                  disp_topo->disp[1].x_origin,
                                  disp_topo->disp[1].y_origin,
                                  disp_topo->disp[0].x_origin,
                                  disp_topo->disp[0].y_origin);
                    temp_disp = disp_topo->disp[0];
                    disp_topo->disp[0] = disp_topo->disp[1];
                    disp_topo->disp[1] = temp_disp;
                    if (disp_edid != NULL)
                    {
                        temp_edid = disp_edid->edid[0];
                        disp_edid->edid[0] = disp_edid->edid[1];
                        disp_edid->edid[1] = temp_edid;
                    }
                }
            }
            else
            {
                mTERA_LOG_ERR("The host SW does not yet support 4 soft client displays");
            }
        }
        //
        // The following code handles the situation where the client provides
        // EDID data. In this case the host SW must reorder the data in the
        // disp_topo and disp_edid structures to match the order of the client
        // EDID data.
        //
        else
        {
            //
            // make a copy of the topology and edid data, this is done to
            // simplify swapping the data in the structures
            //
            sTERA_DISP_TOPO    disp_topo_copy;
            sTERA_DISP_EDID_ID disp_edid_copy[TERA_DISP_MAX_DISP];
            mTERA_MEM_COPY(&disp_topo_copy, disp_topo, sizeof(sTERA_DISP_TOPO));
            mTERA_MEM_COPY(&disp_edid_copy[0], &disp_edid->edid, sizeof(sTERA_DISP_EDID_ID)*TERA_DISP_MAX_DISP);
            mTERA_MEM_ZERO(disp_topo, sizeof(sTERA_DISP_TOPO));
            mTERA_MEM_ZERO(disp_edid, sizeof(sTERA_DISP_EDID));

            // fix problem where the host SW doesn't send the display topology APDU to the FW
            // when the host SW fails to turn on/activate one or more of the attached monitors
            INT32 client_idx;
            for (i=0; i<TERA_DISP_MAX_DISP; i++)
            {
                //
                // Search the disp_edid_copy structure for client entry[i].
                // If a match is found copy the data to the correct location
                // in the disp_edid and disp_topo structures.
                //
                if (edid_match_found(&client_edid->edid[i],
                                     &disp_edid_copy[0],
                                     &client_idx))
                {
                    mTERA_LOG_DBG("sys disp %d matched with client disp %d",
                                  i,
                                  client_idx);
                    disp_topo->disp[i] = disp_topo_copy.disp[client_idx];
                    disp_edid->edid[i] = disp_edid_copy[client_idx];
                }
                else
                {
                    mTERA_LOG_DBG("no match found for sys disp %d in client topo", i);
                }
            }
            for (i=0; i<TERA_DISP_MAX_DISP; i++)
            {
                mTERA_LOG_DBG("  pre-reorder disp%u:  %d x %d @ (%d, %d) rot=%u", i,
                              disp_topo_copy.disp[i].width,
                              disp_topo_copy.disp[i].height,
                              disp_topo_copy.disp[i].x_origin,
                              disp_topo_copy.disp[i].y_origin,
                              disp_topo_copy.disp[i].rotation);
            }
        }

        ret = TERA_SUCCESS;
    } // end if

    // Record the display count
    m_disp_cnt = (ret == TERA_SUCCESS) ? num_disp : 0;

    for (i=0; i<TERA_DISP_MAX_DISP; i++)
    {
        mTERA_LOG_DBG("  post-reorder disp%u: %d x %d @ (%d, %d) rot=%u", i,
                      disp_topo->disp[i].width,
                      disp_topo->disp[i].height,
                      disp_topo->disp[i].x_origin,
                      disp_topo->disp[i].y_origin,
                      disp_topo->disp[i].rotation);
    }

    return ret;
#endif
} // cTERA_DISP::disp_topology_get()

// --------------------------------------------------------------------------
//  cTERA_DISP::disp_cnt_get
//      Return the number of displays
//
//  Description:
//      This function returns the number of displays.
//
INT32 cTERA_DISP::disp_cnt_get(void) const
{
    return m_disp_cnt;
} // cTERA_DISP::disp_cnt_get()

// --------------------------------------------------------------------------
//  cTERA_DISP::check_client_topo_active
//      Return indication of whether the client topo matches the host topo
//
//  Description:
//      This function reads the host topology and compares it to the client
//      display topology. If the topologies match the function returns 1 in
//      the active variable, otherwise the active variable will equal 0. If
//      there is a problem reading the topology the function returns an
//      error code, otherwise the function returns TERA_SUCCESS.
//
INT32 cTERA_DISP::check_client_topo_active(
    sTERA_DISP_EDID_ID *client_disp_edid,
    sTERA_DISP_TOPO    *client_disp_topo,
    INT                *active,
    UINT32             *num_sys_disp
    )
{
#if 1 // TBD - need to implement
    return TERA_SUCCESS;
#else
    DISPLAY_DEVICE      adapter[TERA_DISP_MAX_DISP+1];
    DISPLAY_DEVICE      mon[TERA_DISP_MAX_DISP];
    DEVMODE             mode[TERA_DISP_MAX_DISP];
    sTERA_DISP_EDID_ID  mon_edid_id[TERA_DISP_MAX_DISP];
    INT32               adapter_to_client_mapping[TERA_DISP_MAX_DISP];
    INT32               ret;
    INT                 matched_monitor_cnt=0;
    UINT32              i;
    UINT32              err_mask;
    INT                 client_idx;
    DWORD               client_rot;

    mTERA_LOG_DBG("start");

    for (i=0; i<TERA_DISP_MAX_DISP; i++)
    {
        adapter_to_client_mapping[i] = NO_MAPPING_FOUND;
    }

    ret = host_topology_get(1, adapter, mon, mode, &err_mask, num_sys_disp);
    if (ret == TERA_SUCCESS)
    {
        // Go through the system displays
        for (i=0; i<*num_sys_disp; i++)
        {
            // If a monitor is attached parse its EDID
            if (is_monitor_attached(&mon[i]))
            {
                // Parse the EDID identification data
                if (edid_parse(mon[i].DeviceKey, &mon_edid_id[i]))
                {
                    mTERA_LOG_DBG("  Registry EDID parsed successfully for mon%d", i);

                    //
                    // Determine which monitor in the client topology
                    // maps to this display adapter. If a match is found
                    // the host SW can push the topology settings into the
                    // OS.
                    //
                    if (!edid_match_found(&mon_edid_id[i],
                                          client_disp_edid,
                                          &adapter_to_client_mapping[i]))
                    {
                        mTERA_LOG_ERR("Unable to find a matching client monitor for disp%d!!!",
                                      *num_sys_disp);
                    }
                    else
                    {
                        matched_monitor_cnt++;
                    }
                }
                else
                {
                    err_mask |= TERA_DISP_EDID_PARSE_ERR;
                }
            }
            else
            {
                mTERA_LOG_DBG("  no attached monitor detected!!!");
            }
        }
    }
    // if no matches were found
    if (matched_monitor_cnt == 0)
    {
        //
        // If there is only one system adapter and one client display
        // the host SW should try to push the client settings to the host.
        // This helps when the hard host filters the monitor hotplug signal.
        // In this situation it's better to try to push the client settings
        // rather than do nothing.
        //
        if ((1 == *num_sys_disp) && (1 == client_disp_topo->num_disp))
        {
            adapter_to_client_mapping[0] = 0;
        }
        else
        {
            err_mask |= TERA_DISP_EDID_MATCH_NOT_FOUND;
        }
    }

    // check if the client and system adapter topologies are equal
    *active = 0;
    if ((ret == TERA_SUCCESS) &&
        (err_mask == 0) &&
        (*num_sys_disp >= client_disp_topo->num_disp))
    {
        // walk through the system adapters
        *active = 1;
        for (i=0; i<*num_sys_disp; i++)
        {
            client_idx = adapter_to_client_mapping[i];
            if (client_idx != NO_MAPPING_FOUND)
            {
                //
                // convert the client orientation to the same type of variable
                // as the system orientation
                //
                client_rot = (client_disp_topo->disp[client_idx].rotation == TERA_DISP_ROTATION_0) ?
                              DMDO_DEFAULT :
                             (client_disp_topo->disp[client_idx].rotation == TERA_DISP_ROTATION_90) ?
                              DMDO_90 :
                             (client_disp_topo->disp[client_idx].rotation == TERA_DISP_ROTATION_180) ?
                              DMDO_180 : DMDO_270;

                //
                // if the rotation is 90 or 270 the client height and width
                // need to be swapped. This must be done because the FW reports
                // pre-rotated values
                //
                if ((client_rot == DMDO_90) || (client_rot == DMDO_270))
                {
                    DWORD temp_width;
                    temp_width = mode[i].dmPelsWidth;
                    mode[i].dmPelsWidth = mode[i].dmPelsHeight;
                    mode[i].dmPelsHeight = temp_width;
                }

                //
                // if either position field is non-zero the positions should
                // be identical
                //
                if (mode[i].dmPosition.x || mode[i].dmPosition.y)
                {
                    if ((mode[i].dmPelsWidth != client_disp_topo->disp[client_idx].width) ||
                        (mode[i].dmPelsHeight != client_disp_topo->disp[client_idx].height) ||
                        (mode[i].dmPosition.x != client_disp_topo->disp[client_idx].x_origin) ||
                        (mode[i].dmPosition.y != client_disp_topo->disp[client_idx].y_origin) ||
                        (mode[i].dmDisplayOrientation != client_rot) ||
                        (!mTERA_REFRESH_RATES_SAME(mode[i].dmDisplayFrequency,
                                                   (DWORD)client_disp_topo->disp[client_idx].refresh_rate)))
                    {
                        *active = 0;
                        break;
                    }
                }
                // if both position fields are 0 compare everything but the
                // position, this deals with the scenario where the client
                // supplies topology data for 2 monitors but only one is
                // connected and the one that is connected does not use the
                // position (0,0)
                else
                {
                    if ((mode[i].dmPelsWidth != client_disp_topo->disp[client_idx].width) ||
                        (mode[i].dmPelsHeight != client_disp_topo->disp[client_idx].height) ||
                        (mode[i].dmDisplayOrientation != client_rot) ||
                        (!mTERA_REFRESH_RATES_SAME(mode[i].dmDisplayFrequency,
                                                   (DWORD)client_disp_topo->disp[client_idx].refresh_rate)))
                    {
                        *active = 0;
                        break;
                    }
                }
            }
            else
            {
                mTERA_LOG_ERR("Client monitor not found for system adapter %d", i);
                *active = 0;
                break;
            }
        }
        if (*active == 0)
        {
            if (client_idx != NO_MAPPING_FOUND)
            {
                mTERA_LOG_DBG("Topo differ: Client topo%d %d x %d @ (%d, %d) rot=%d freq=%d",
                              client_idx,
                              client_disp_topo->disp[client_idx].width,
                              client_disp_topo->disp[client_idx].height,
                              client_disp_topo->disp[client_idx].x_origin,
                              client_disp_topo->disp[client_idx].y_origin,
                              client_rot,
                              client_disp_topo->disp[client_idx].refresh_rate);
            }
            mTERA_LOG_DBG("             System topo%d %d x %d @ (%d, %d) rot=%d freq=%d",
                          i,
                          mode[i].dmPelsWidth,
                          mode[i].dmPelsHeight,
                          mode[i].dmPosition.x,
                          mode[i].dmPosition.y,
                          mode[i].dmDisplayOrientation,
                          mode[i].dmDisplayFrequency);
        }
        else
        {
            // verify the number of active displays equals the number of matched
            // monitors
            UINT32              num_active_disp;
            num_active_disp = GetSystemMetrics(SM_CMONITORS);
            if (num_active_disp != *num_sys_disp)
            {
                mTERA_LOG_DBG("More active display adapters (%d) than matched client displays (%d)",
                              num_active_disp,
                              *num_sys_disp);
                *active = 0;
            }
            else
            {
                mTERA_LOG_DBG("Topologies identical!!!");
            }
        }
        return TERA_SUCCESS;
    }
    mTERA_LOG_DBG("Not comparing topologies: ret=%d, err_mask=0x%08X, num_sys_disp=%d, client topo->num_disp=%d",
                  ret, err_mask, *num_sys_disp, client_disp_topo->num_disp);
    if ((ret == TERA_SUCCESS) && (err_mask == 0))
    {
        return TERA_SUCCESS;
    }
    else
    {
        return TERA_ERR_FAILURE;
    }
#endif
} // cTERA_DISP::check_client_topo_active()

// --------------------------------------------------------------------------
//  cTERA_DISP::push_topology_to_os
//      Push client topology to the host operating system
//
//  Description:
//      This function pushes the display topology settings received from the
//      client into the host operating system.
//
INT32 cTERA_DISP::push_topology_to_os(
    sTERA_DISP_TOPO    *client_disp_topo,
    sTERA_DISP_EDID_ID *client_disp_edid,
    UINT32             *err_mask
    )
{
#if 1 // TBD - need to implement
    return TERA_SUCCESS;
#else
    DISPLAY_DEVICE      adapter[TERA_DISP_MAX_DISP+1];
    DISPLAY_DEVICE      mon[TERA_DISP_MAX_DISP];
    DEVMODE             mode[TERA_DISP_MAX_DISP];
    DEVMODE             old_mode[TERA_DISP_MAX_DISP];
    sTERA_DISP_EDID_ID  mon_edid_id[TERA_DISP_MAX_DISP];
    INT32               adapter_to_client_mapping[TERA_DISP_MAX_DISP];
    INT32               ret;
    UINT32              i;
    UINT32              num_sys_disp;
    LONG                err;
    UINT32              client_disp;
    UINT32              sys_disp;

    mTERA_LOG_DBG("start");

    for (i=0; i<TERA_DISP_MAX_DISP; i++)
    {
        adapter_to_client_mapping[i] = NO_MAPPING_FOUND;
    }

    ret = host_topology_get(0, adapter, mon, mode, err_mask, &num_sys_disp);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("  ERROR getting host topology, not pushing client topology (ret=%d)", ret);
        return ret;
    }

    // Go through the system displays
    for (i=0; i<TERA_DISP_MAX_DISP; i++)
    {
        // If an adapter is active without an attached monitor
        if (is_adapter_active_no_monitor(&mon[i]))
        {
            dump_display_device(&mon[i]);

            mTERA_LOG_DBG("  adapter active but no attached monitor detected for sys disp %d!!!", i);
            adapter_to_client_mapping[i] = ADAPTER_WITHOUT_MONITOR;
        }
        // If a monitor is attached parse its EDID
        else if (is_monitor_attached(&mon[i]))
        {
            dump_display_device(&mon[i]);

            // Parse the attached monitor EDID
            if (edid_parse(mon[i].DeviceKey, &mon_edid_id[i]))
            {
                mTERA_LOG_DBG("  Successfully parsed registry EDID for mon%d", i);

                //
                // Determine which monitor in the client topology
                // maps to this display adapter. If a match is found
                // the host SW can push the topology settings into the
                // OS.
                //
                if (!edid_match_found(&mon_edid_id[i],
                                      client_disp_edid,
                                      &adapter_to_client_mapping[i]))
                {
                    mTERA_LOG_ERR("Unable to find a matching client monitor for disp%d!!!",
                                  num_sys_disp);
                    *err_mask |= TERA_DISP_EDID_MATCH_NOT_FOUND;
                }
            }
            else
            {
                *err_mask |= TERA_DISP_EDID_PARSE_ERR;
            }
        }
        else
        {
            mTERA_LOG_DBG("  mon%d not attached or active", i);
        }
    }

    if (num_sys_disp == 0)
    {
        mTERA_LOG_ERR("End - no active adapters detected!!!");
        *err_mask |= TERA_DISP_ERR_NO_SYS_DISP;
        return TERA_ERR_FAILURE;
    }
    //
    // Its okay if a matching EDID was not found, this can happen if the host
    // reports 2 active displays and the client topology only contains one.
    // This really should not happen, but I've seen it on some host cards that
    // have HW monitor emulation and present monitor information from a
    // previous session. The HC10 is one system that does this.
    //
    if ((*err_mask != 0) &&
        (*err_mask != TERA_DISP_EDID_MATCH_NOT_FOUND) &&
        (*err_mask != TERA_DISP_EDID_PARSE_ERR))
    {
        mTERA_LOG_ERR("None zero err_mask=%08X", *err_mask);
        return TERA_ERR_FAILURE;
    }

    mTERA_LOG_DBG("adapter_to_client_mapping[%d, %d, %d, %d]",
                  adapter_to_client_mapping[0],
                  adapter_to_client_mapping[1],
                  adapter_to_client_mapping[2],
                  adapter_to_client_mapping[3]);
    if (!is_client_mapping_valid(adapter_to_client_mapping, mode))
    {
        mTERA_LOG_ERR("BAD mapping in adapter_to_client_mapping");
        return TERA_ERR_FAILURE;
    }

    //
    // Make a copy of the current mode settings. This is used to determine
    // if changes are made to the settings. If no changes are made the code
    // doesn't write the settings to the host. This avoids screen blanking.
    //
    mTERA_MEM_COPY(&old_mode[0], &mode[0], sizeof(DEVMODE)*TERA_DISP_MAX_DISP);

    // Prepare the variables needed to push the settings to the host
    for (sys_disp=0; sys_disp<num_sys_disp; sys_disp++)
    {
        //
        // Update the display adapter settings using the client topology. If
        // a match for this display adapter was not found the monitor might
        // need to be turned on.
        //
        if (adapter_to_client_mapping[sys_disp] == NO_MAPPING_FOUND)
        {
            mTERA_LOG_DBG("unable to find a mapping for system display %d",
                          sys_disp);

            //
            // The following code only works on dual monitor system. The
            // code maps the monitor with no mapping to the unmapped
            // display adapter.
            //
            if (num_sys_disp == 2)
            {
                if (adapter_to_client_mapping[0] == NO_MAPPING_FOUND)
                {
                    adapter_to_client_mapping[0] = (adapter_to_client_mapping[1] + 1) % 2;
                }
                else
                {
                    adapter_to_client_mapping[1] = (adapter_to_client_mapping[0] + 1) % 2;
                }
            }
        }
        if ((adapter_to_client_mapping[sys_disp] != NO_MAPPING_FOUND) &&
            (adapter_to_client_mapping[sys_disp] != ADAPTER_WITHOUT_MONITOR))
        {
            client_disp = adapter_to_client_mapping[sys_disp];

            //
            // Configure the width and height settings. If a match for this display
            // adapter was not found and the client topology only contains
            // information for 1 monitor the display should be turned off. This is
            // done by setting the height and width equal to zero.
            //
            if((client_disp_topo->disp[client_disp].rotation == TERA_DISP_ROTATION_0) ||
               (client_disp_topo->disp[client_disp].rotation == TERA_DISP_ROTATION_180))
            {
                mode[sys_disp].dmPelsWidth  = client_disp_topo->disp[client_disp].width;
                mode[sys_disp].dmPelsHeight = client_disp_topo->disp[client_disp].height;
            }
            else
            {
                mode[sys_disp].dmPelsWidth  = client_disp_topo->disp[client_disp].height;
                mode[sys_disp].dmPelsHeight = client_disp_topo->disp[client_disp].width;
            }

            // configure the position settings
            if (num_sys_disp == 1)
            {
                //
                // Override the client position settings if only one monitor is
                // is active. The position might be incorrect if the client has
                // configured multiple displays but only one monitor is
                // connected.
                //
                mode[sys_disp].dmPosition.x = 0;
                mode[sys_disp].dmPosition.y = 0;
            }
            else
            {
                mode[sys_disp].dmPosition.x = client_disp_topo->disp[client_disp].x_origin;
                mode[sys_disp].dmPosition.y = client_disp_topo->disp[client_disp].y_origin;
            }

            // configure the rotation settings
            mode[sys_disp].dmDisplayOrientation =
                    (client_disp_topo->disp[client_disp].rotation == TERA_DISP_ROTATION_0) ?
                    DMDO_DEFAULT :
                    (client_disp_topo->disp[client_disp].rotation == TERA_DISP_ROTATION_90) ?
                    DMDO_90 :
                    (client_disp_topo->disp[client_disp].rotation == TERA_DISP_ROTATION_180) ?
                    DMDO_180 : DMDO_270;

            // configure the frequency setting
            mode[sys_disp].dmDisplayFrequency = client_disp_topo->disp[client_disp].refresh_rate;

            //
            // Check the fixed output setting (unfortunately this cannot be
            // changed. If it is not = DMDFO_DEFAULT the settings will not
            // be applied successfully, and in some situations the call to
            // ChangeDisplaySettingsEx() may hang indefinitely. This problem has
            // been observed on the ATI wishbringer while connecting different
            // clients with different sized monitors. The driver reverts to a
            // scaled mode sometimes when connecting a client with a smaller
            // monitor than the monitor of the previously connected client.
            //
            if ((mode[sys_disp].dmFields & DM_DISPLAYFIXEDOUTPUT) &&
                (mode[sys_disp].dmDisplayFixedOutput != DMDFO_DEFAULT))
            {
                mTERA_LOG_ERR("dmDisplayFixedOutput=%d != DMDFO_DEFAULT=%d, not updating host settings!!!",
                              mode[sys_disp].dmDisplayFixedOutput,
                              DMDFO_DEFAULT);

                return TERA_ERR_INVALID_STATE;
            }

            // Activate the monitor if the width and height are non zero
            if ((client_disp_topo->disp[client_disp].width) &&
                (client_disp_topo->disp[client_disp].height))
            {
                mode[sys_disp].dmFields = mode[sys_disp].dmFields | DM_POSITION;
            }
        }
        else
        {
            mTERA_LOG_DBG("No mapping found for adapter %d, turning off adapter",
                          sys_disp);
            mode[sys_disp].dmPelsWidth = 0;
            mode[sys_disp].dmPelsHeight = 0;
        }
    }

    // exit if the mode settings don't change
    if (0 == memcmp(&old_mode[0], &mode[0], sizeof(DEVMODE)*TERA_DISP_MAX_DISP))
    {
        mTERA_LOG_DBG("Display settings not updated because nothing changed");
        err = DISP_CHANGE_SUCCESSFUL;
    }
    else
    {
        // dump the old and new settings
        UINT32 max_disp;
        if (client_disp_topo->num_disp >= num_sys_disp)
        {
            max_disp = client_disp_topo->num_disp;
        }
        else
        {
            max_disp = num_sys_disp;
        }
        for (i=0; i<max_disp; i++)
        {
            mTERA_LOG_DBG("Old settings display%d", i);
            dump_display_mode(&old_mode[i]);
            mTERA_LOG_DBG("New settings display%d", i);
            dump_display_mode(&mode[i]);
        }

        //
        // Configure display 1 first when the following conditions are true:
        // - Display 0 is to be turned off. This fixes a problem seen with the
        //   ATI driver where it automatically enables the 2nd display if the
        //   first and only display is turned off
        // - Display 1 is the primary display and 2 displays are active
        //
        INT first_mon = 0;
        if (((mode[0].dmPelsWidth == 0) &&
             (mode[0].dmPelsHeight == 0)) ||
            ((num_sys_disp == 2) &&
             (mode[1].dmPosition.x == 0) &&
             (mode[1].dmPosition.y == 0)))
        {
            first_mon = 1;
        }
        INT cfg_mon = first_mon;

        //
        // Push the display settings to the OS for each adapter that was matched
        // with a display.
        //
        for (sys_disp=0; sys_disp<num_sys_disp; sys_disp++)
        {
            //
            // Update the settings in the registry but don't activate them
            // yet. The new settings are activated after this for loop if all
            // settings are applied successfuly.
            //
            ret = update_display_setting(cfg_mon,
                                         CDS_UPDATEREGISTRY | CDS_NORESET,
                                         &mode[cfg_mon],
                                         &adapter[cfg_mon],
                                         err_mask,
                                         &err);

            // give up if there was an error and the display was active
            if (*err_mask & TERA_DISP_ERR_CHANGING_OS_SETTINGS)
            {
                mTERA_LOG_DBG("TERA_DISP_ERR_CHANGING_OS_SETTINGS set in err_mask (%d)",
                              *err_mask);
                break;
            }

            cfg_mon = (cfg_mon + 1) % num_sys_disp;
        }

        // activate new settings if all displays were updated successfully
        if(err == DISP_CHANGE_SUCCESSFUL)
        {
            mTERA_LOG_DBG("Activate new display settings");
            err = ChangeDisplaySettingsEx(NULL, NULL, NULL, CDS_RESET, NULL);
            change_settings_debug(err);
            if(err != DISP_CHANGE_SUCCESSFUL)
            {
                *err_mask |= TERA_DISP_ERR_CHANGING_OS_SETTINGS;
            }
        }

        //
        // If there was a problem updating the display settings try updating
        // the settings for each individual display
        //
        if(err != DISP_CHANGE_SUCCESSFUL)
        {
            cfg_mon = first_mon;
            for (sys_disp=0; sys_disp<num_sys_disp; sys_disp++)
            {
                // Only update the settings if a resolution, orientation or
                // frequency setting(s) changed. Don't check the position
                // setting, because this can trigger the screen blanking
                // problem (bug 7437).
                if ((old_mode[cfg_mon].dmPelsWidth != mode[cfg_mon].dmPelsWidth) ||
                    (old_mode[cfg_mon].dmPelsHeight != mode[cfg_mon].dmPelsHeight) ||
                    (old_mode[cfg_mon].dmDisplayOrientation != mode[cfg_mon].dmDisplayOrientation) ||
                    (old_mode[cfg_mon].dmDisplayFrequency != mode[cfg_mon].dmDisplayFrequency))
                {
                    ret = update_display_setting(cfg_mon,
                                                 CDS_UPDATEREGISTRY | CDS_RESET,
                                                 &mode[cfg_mon],
                                                 &adapter[cfg_mon],
                                                 err_mask,
                                                 &err);
                }
                else
                {
                    mTERA_LOG_DBG("Not updating individual display%d settings because they're already in use",
                                  cfg_mon);
                }

                cfg_mon = (cfg_mon + 1) % num_sys_disp;
            }
        }
    }

    if(err == DISP_CHANGE_SUCCESSFUL)
    {
        ret = TERA_SUCCESS;
    }
    else
    {
        mTERA_LOG_DBG("No display settings were changed: num client displays=%d, num attached displays=%d",
                      client_disp_topo->num_disp, num_sys_disp);
        ret = TERA_ERR_FAILURE;
    }
    mTERA_LOG_DBG("end - ret=%d", ret);
    return ret;
#endif
} // cTERA_DISP::push_topology_to_os()

// --------------------------------------------------------------------------
//  cTERA_DISP::edid_data_unique
//      Look through the client cached topology for duplicate EDID data
//
//  Description:
//      This function walks through the client cached display topology EDID
//      data and checks to see if any entries are the same. If a duplicate
//      entry is found the function returns 0, otherwise it returns 1. This
//      should only happen if two monitors are connected to the client and
//      the EDID information is not unique.
//
//      NOTE: this problem has been observed in a lab at IBM, which is why the
//      check is included. The problem can also happen on the HC10 when monitor
//      emulation is active. I've also seen it happen when the soft client is
//      connected to a HH.
//
INT cTERA_DISP::edid_data_unique(
    sTERA_DISP_EDID_ID  *client_disp_edid,
    INT32               num_disp
    )
{
    INT i, j;

    // this check only needs to be performed if there is more than 1 display
    if (num_disp > 1)
    {
        // loop through the client topology entries searching for duplicate data
        for (i=0; i<(TERA_DISP_MAX_DISP-1); i++)
        {
            // skip this entry if the VID and PID strings are empty
            if ((strlen(client_disp_edid[i].vid) != 0) &&
                (strlen(client_disp_edid[i].pid) != 0))
            {
                for (j=i+1; j<TERA_DISP_MAX_DISP; j++)
                {
                    // walk through the other EDID entries looking for a match
                    if (!strncmp(client_disp_edid[i].vid,    client_disp_edid[j].vid,    TERA_MON_EDID_VID_LEN) &&
                        !strncmp(client_disp_edid[i].pid,    client_disp_edid[j].pid,    TERA_MON_EDID_PID_LEN) &&
                        !strncmp(client_disp_edid[i].serial, client_disp_edid[j].serial, TERA_MON_EDID_SERIAL_LEN))
                    {
                        mTERA_LOG_DBG("Duplicate EDID entries (%d, %d) found in cached topo", i, j);
                        return 0;
                    }
                }
            }
        }
    }

    mTERA_LOG_DBG("edid data is unique");
    return 1;

} // cTERA_DISP::edid_data_unique()

// --------------------------------------------------------------------------
//  cTERA_DISP::cached_topo_is_valid
//      Look through the client cached topology for valid data
//
//  Description:
//      This function walks through the client cached display topology data
//      and checks to see if any entries are not blank. If a non blank entry is
//      found the function returns 1, otherwise it returns 0.
//
INT cTERA_DISP::cached_topo_is_valid(
    sTERA_DISP_TOPO     *client_disp_topo
    )
{
    INT i;

    // loop through the client topology entries searching for a non blank entry
    for (i=0; i<TERA_DISP_MAX_DISP; i++)
    {
        //
        // valid topology is present if any displays width and height are
        // non-zero
        //
        if ((client_disp_topo->disp[i].width) &&
            (client_disp_topo->disp[i].height))
        {
            mTERA_LOG_DBG("TRUE: valid topology data found for disp%d", i);
            return 1;
        }
    }

    mTERA_LOG_DBG("FALSE: client cached topology is empty");
    return 0;

} // cTERA_DISP::cached_topo_is_valid()

// --------------------------------------------------------------------------
//  cTERA_DISP::monitor_rescan
//      Trigger the device manager to scan for new monitors
//
//  Description:
//      This function triggers the device manager to scan for new monitors. This
//      function is useful for detecting new monitors that are not yet detected
//      or deleted from the device manager manually.
//
void cTERA_DISP::monitor_rescan(void)
{
#if 0 // TBD - need to implement
//
// Note: While writing this function I found a couple ways to implement it.
//       Both versions are included but one is commented out. I tested
//       both hoping one of them was not privileged, requiring UAC to be
//       disabled, but that was not the case. The unused version could be
//       deleted but I decided to include it in case there might be a use for
//       it in the future.
//
#if 1
    DEVINST root;

    if (CM_Locate_DevNode(&root, NULL, CM_LOCATE_DEVNODE_NORMAL) !=
        CR_SUCCESS)
    {
        mTERA_LOG_ERR("CM_Locate_DevNode failed");
    }
    else
    {
        if (CM_Reenumerate_DevNode(root, CM_REENUMERATE_NORMAL) != CR_SUCCESS)
        {
            mTERA_LOG_ERR("CM_Reenumerate_DevNode failed");
        }
        else
        {
            mTERA_LOG_DBG("scanning for new hardware succeeded");
        }
    }
#else
    HDEVINFO hDevInfo;
    SP_DEVINFO_DATA DeviceInfoData;
    DWORD i;

    mTERA_LOG_DBG("start");

    // Create a HDEVINFO with all present monitors
    hDevInfo = SetupDiGetClassDevs(&GUID_DEVINTERFACE_MONITOR,
                                   0, // Enumerator
                                   0,
                                   DIGCF_PRESENT | DIGCF_DEVICEINTERFACE );

    if (hDevInfo == INVALID_HANDLE_VALUE)
    {
        // Insert error handling here.
        mTERA_LOG_ERR("SetupDiGetClassDevs returned invalid handle");
    }

    // Enumerate through all devices in Set.
    DeviceInfoData.cbSize = sizeof(SP_DEVINFO_DATA);
    for (i=0;
         SetupDiEnumDeviceInfo(hDevInfo, i, &DeviceInfoData);
         i++)
    {
        DWORD DataT;
        LPTSTR buffer = NULL;
        DWORD buffersize = 0;

        //
        // Call function with null to begin with,
        // then use the returned buffer size (doubled)
        // to Alloc the buffer. Keep calling until
        // success or an unknown failure.
        //
        // Double the returned buffersize to correct
        // for underlying legacy CM functions that
        // return an incorrect buffersize value on
        // DBCS/MBCS systems.
        //
        while (!SetupDiGetDeviceRegistryProperty(hDevInfo,
                                                 &DeviceInfoData,
                                                 //SPDRP_HARDWAREID,
                                                 SPDRP_DEVICEDESC,
                                                 &DataT,
                                                 (PBYTE)buffer,
                                                 buffersize,
                                                 &buffersize))
        {
            if (GetLastError() == ERROR_INSUFFICIENT_BUFFER)
            {
                // Change the buffer size.
                if (buffer) mTERA_MEM_FREE(buffer);
                // Double the size to avoid problems on
                // W2k MBCS systems per KB 888609.
                buffer = (LPTSTR)mTERA_MEM_ALLOC(buffersize * 2);
            }
            else
            {
                // Insert error handling here.
                mTERA_LOG_ERR("GetLastError != INSUFFICIENT_BUFFER (%u)",
                               GetLastError());
                break;
            }
        }

        mTERA_LOG_DBG("Result:[%s]", buffer);

        if (buffer)
        {
            mTERA_MEM_FREE(buffer);
        }
    }

    if ((GetLastError() != NO_ERROR) &&
        (GetLastError() != ERROR_NO_MORE_ITEMS))
    {
        // Insert error handling here.
        mTERA_LOG_ERR("GetLastError = %u", GetLastError());
    }

    //  Cleanup
    SetupDiDestroyDeviceInfoList(hDevInfo);

#endif
#endif
} // cTERA_DISP::monitor_rescan()

// --------------------------------------------------------------------------
//  update_display_setting
//      Update the configuration settings for a display
//
//  Description:
//      This function writes new topology settings for a single display to
//      the host operating system. The caller specifies most of the parameters
//      supplied to the Windows function that updates the display settings
//      (ChangeDisplaySettingsEx).
//
//      By allowing the caller to provide the flags parameter this function can
//      be used to activate new settings immediately or it can be used to
//      update the settings in the registry. This allows the caller to later
//      activate the settings in the registry. This is useful when updating the
//      settings for multiple displays. In this scenario you should update all
//      of the registry settings and then activate the settings for the
//      displays at the same time.
//
#if 0 // TBD - need to implement
INT32 cTERA_DISP::update_display_setting(
    INT             cfg_mon,
    DWORD           flags,
    DEVMODE        *mode,
    DISPLAY_DEVICE *adapter,
    UINT32         *err_mask,
    LONG           *err
    )
{
    INT32   ret;

    // don't update the settings for this display if it is off
    if (0 == (mode->dmFields & DM_POSITION))
    {
        mTERA_LOG_DBG("Not applying new settings to disp%d because display is already OFF", cfg_mon);
        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else
    {
        //
        // The CDS_SET_PRIMARY flag is set for the display located at position
        // (0,0). This is consistent with what Windows does. The client
        // provides a field in the cached topology message that indicates
        // which display is the primary, but the host SW does not use this
        // field.
        //
        if ((mode->dmPosition.x == 0) &&
            (mode->dmPosition.y == 0) &&
            (mode->dmPelsWidth != 0) &&
            (mode->dmPelsHeight != 0))
        {
            flags |= CDS_SET_PRIMARY;
        }

        INT retry_cnt=0;
        do
        {
            *err = ChangeDisplaySettingsEx(adapter->DeviceName,
                                           mode,
                                           NULL,
                                           flags,
                                           NULL);
            change_settings_debug(*err);

            //
            // If there was an error and the display is being turned off try to
            // turn it off by providing a NULL mode structure. This sometimes
            // works with the NVIDIA 8600 GT GPU (driver version 196.21).
            //
            if ((*err == DISP_CHANGE_BADMODE) &&
                ((mode->dmPelsWidth == 0) &&
                 (mode->dmPelsHeight == 0)))
            {
                mTERA_LOG_DBG("disabling disp%d using NULL mode", cfg_mon);
                *err = ChangeDisplaySettingsEx(adapter->DeviceName,
                                               NULL,
                                               NULL,
                                               flags,
                                               NULL);
                change_settings_debug(*err);
            }
            retry_cnt++;
        } while ((retry_cnt < 5) && (*err != DISP_CHANGE_SUCCESSFUL));

        // give up if there was an error and the display was active
        if((*err != DISP_CHANGE_SUCCESSFUL) &&
           ((mode->dmPelsWidth != 0) && (mode->dmPelsHeight != 0)))
        {
            *err_mask |= TERA_DISP_ERR_CHANGING_OS_SETTINGS;
            ret = TERA_ERR_FAILURE;
        }
        else
        {
            ret = TERA_SUCCESS;
        }
    }

    mTERA_LOG_DBG("end: disp%d, err_mask=%d, ret=%d, err=%d",
                  cfg_mon,
                  *err_mask,
                  ret,
                  *err);
    return ret;
} // update_display_setting()
#endif

// --------------------------------------------------------------------------
//  cTERA_DISP::host_topology_get
//      Read the host topology
//
//  Description:
//      This function reads the host display topology settings and returns
//      host topology structures. This is a utility function used within
//      the TERA_DISP module.
//
#if 0 // TBD - need to implement
INT32 cTERA_DISP::host_topology_get(
    INT                 active_disp,    // when non-zero only report active displays
    DISPLAY_DEVICE     *adapter,
    DISPLAY_DEVICE     *mon,
    DEVMODE            *mode,
    UINT32             *err_mask,
    UINT32             *num_sys_disp
    )
{
    UINT32              i = 0;
    UINT32              j;
    INT                 disp_settings_found;
    INT                 monitor_found;
    DISPLAY_DEVICE     *start_mon = mon;
    DEVMODE            *start_mode = mode;

    mTERA_LOG_DBG("start: active_disp=%d", active_disp);

    mTERA_MEM_ZERO(adapter, sizeof(DISPLAY_DEVICE));
    adapter->cb = sizeof(DISPLAY_DEVICE);
    *err_mask = 0;
    *num_sys_disp = 0;

    //
    // Go through all the display adapters and get the display settings for
    // each display adapter
    //
    while (EnumDisplayDevices(NULL, i++, adapter, 0))
    {
        dump_display_device(adapter);

        // if the monitor connected to the display adapter is not a pseudo device
        if (!(adapter->StateFlags & DISPLAY_DEVICE_MIRRORING_DRIVER) &&
             ((!active_disp) || (adapter->StateFlags & DISPLAY_DEVICE_ATTACHED_TO_DESKTOP)))
        {
            mTERA_LOG_DBG(" attached monitor!!!");

            disp_settings_found = 1;
            mTERA_MEM_ZERO(mode, sizeof(DEVMODE));
            mode->dmSize = sizeof(DEVMODE);

            //
            // Get the display setting from cache. This can fail if the display
            // is not active.
            //
            if (!EnumDisplaySettings(adapter->DeviceName,
                                     ENUM_CURRENT_SETTINGS,
                                     mode))
            {
                mTERA_LOG_DBG("Unable to get display settings from the cache (ret=%u), trying registry",
                              GetLastError());

                // Try to get the display settings from the registry
                if (!EnumDisplaySettings(adapter->DeviceName,
                                         ENUM_REGISTRY_SETTINGS,
                                         mode))
                {
                    mTERA_LOG_ERR("failed to get display settings from registry %u",
                                  GetLastError());
                    disp_settings_found = 0;
                    *err_mask |= TERA_DISP_HOST_EDID_NOT_FOUND;
                }
            }

            // Look for the monitor attached to the adapter
            mTERA_MEM_ZERO(mon, sizeof(DISPLAY_DEVICE));
            mon->cb = sizeof(DISPLAY_DEVICE);
            j = 0;
            monitor_found = 0;
            while (EnumDisplayDevices(adapter->DeviceName,
                   j++, mon, EDD_GET_DEVICE_INTERFACE_NAME))
            {
                mTERA_LOG_DBG(" mon.StateFlags = 0x%08X, len DeviceID=%d, len DeviceString=%d, j=%d",
                              mon->StateFlags, wcslen(mon->DeviceID), wcslen(mon->DeviceString), j);

                // Verify a monitor is attached
                if (is_monitor_attached(mon))
                {
                    dump_display_device(mon);
                    monitor_found = 1;
                    break;
                }
                else
                {
                    mTERA_LOG_DBG("  no attached monitor detected!!!");
                }
            }

            //
            // If the display was found AND the required DEVMODE fields are
            // initialized. The DM_POSITION field is not set if the monitor
            // is not active.
            //
            if ((disp_settings_found) &&
                (mode->dmFields & DM_PELSWIDTH) &&
                (mode->dmFields & DM_PELSHEIGHT) &&
                (mode->dmFields & DM_DISPLAYORIENTATION) &&
                (mode->dmFields & DM_DISPLAYFREQUENCY))
            {
                *num_sys_disp = *num_sys_disp + 1;
                adapter++;
                mon++;
                mode++;
            }
            // Disabled monitor or the required fields are not supported
            else
            {
                mTERA_LOG_DBG("monitor disabled, not attached or a mode is not supported (dmFields=0x%08X)",
                               mode->dmFields);
                if(!(mon->StateFlags & DISPLAY_DEVICE_ATTACHED))
                {
                    mTERA_LOG_DBG("  No monitor attached to display adapter");
                }
                if(!(mode->dmFields & DM_PELSWIDTH))
                {
                    mTERA_LOG_ERR("  DM_PELSWIDTH in dmFields is not set");
                }
                if(!(mode->dmFields & DM_PELSHEIGHT))
                {
                    mTERA_LOG_ERR("  DM_PELSHEIGHT in dmFields is not set");
                }
                if(!(mode->dmFields & DM_DISPLAYORIENTATION))
                {
                    mTERA_LOG_ERR("  DM_DISPLAYORIENTATION in dmFields is not set");
                }
                if(!(mode->dmFields & DM_DISPLAYFREQUENCY))
                {
                    mTERA_LOG_ERR("  DM_DISPLAYFREQUENCY in dmFields is not set");
                }
                if(!disp_settings_found)
                {
                    *err_mask |= TERA_DISP_ERR_NO_SYS_DISP;
                    mTERA_LOG_ERR("problem retrieving display settings: num_sys_disp=%d, err_mask=0x%08X",
                                  *num_sys_disp, *err_mask);
                    return TERA_ERR_FAILURE;
                }
                if(!monitor_found)
                {
                    mTERA_LOG_ERR("problem finding monitor or reading the EDID data");
                }
            }
        }
        else
        {
            // Ignore display adapters not used in the desktop
            mTERA_LOG_DBG("  ignoring unused display (%s)",
                          adapter->DeviceString);
        }

        if (*num_sys_disp <= TERA_DISP_MAX_DISP)
        {
            mTERA_MEM_ZERO(adapter, sizeof(DISPLAY_DEVICE));
            adapter->cb = sizeof(DISPLAY_DEVICE);
        }
        else
        {
            *err_mask |= TERA_DISP_ERR_TOO_MANY_SYS_DISP;
            mTERA_LOG_ERR("End - too many active adapters (%d) detected: err_mask=0x%08X",
                          *num_sys_disp, *err_mask);
            return TERA_ERR_FAILURE;
        }
    } // end while

    mTERA_LOG_DBG("end: num_sys_disp=%d, err_mask=0x%08X",
                  *num_sys_disp, *err_mask);

    // Go through the system displays
    for (i=0; i<*num_sys_disp; i++)
    {
        dump_display_device(start_mon);
        dump_display_mode(start_mode);
        start_mon++;
        start_mode++;
    }
    return TERA_SUCCESS;

} // cTERA_DISP::host_topology_get()
#endif

// --------------------------------------------------------------------------
//  cTERA_DISP::change_settings_debug
//      Generate debug message based on ChangeDisplaySettingsEx return code
//
//  Description:
//      This function interprets the return code from the
//      ChangeDisplaySettingsEx function and outputs a debug message.
//
#if 0 // TBD - need to implement
void cTERA_DISP::change_settings_debug(
        LONG ret
    )
{
    switch(ret)
    {
        case DISP_CHANGE_SUCCESSFUL:
            mTERA_LOG_DBG("ChangeDisplaySettingsEx returned DISP_CHANGE_SUCCESSFUL");
            break;
        case DISP_CHANGE_BADDUALVIEW:
            mTERA_LOG_DBG("ChangeDisplaySettingsEx returned DISP_CHANGE_BADDUALVIEW");
            break;
        case DISP_CHANGE_BADFLAGS:
            mTERA_LOG_DBG("ChangeDisplaySettingsEx returned DISP_CHANGE_BADFLAGS");
            break;
        case DISP_CHANGE_BADMODE:
            mTERA_LOG_DBG("ChangeDisplaySettingsEx returned DISP_CHANGE_BADMODE");
            break;
        case DISP_CHANGE_BADPARAM:
            mTERA_LOG_DBG("ChangeDisplaySettingsEx returned DISP_CHANGE_BADPARAM");
            break;
        case DISP_CHANGE_FAILED:
            mTERA_LOG_DBG("ChangeDisplaySettingsEx returned DISP_CHANGE_FAILED");
            break;
        case DISP_CHANGE_NOTUPDATED:
            mTERA_LOG_DBG("ChangeDisplaySettingsEx returned DISP_CHANGE_NOTUPDATED");
            break;
        case DISP_CHANGE_RESTART:
            mTERA_LOG_DBG("ChangeDisplaySettingsEx returned DISP_CHANGE_RESTART");
            break;
        default:
            mTERA_LOG_DBG("ChangeDisplaySettingsEx returned unexpected ret=%d", ret);
            break;
    }
} // cTERA_DISP::change_settings_debug()
#endif

// --------------------------------------------------------------------------
//  cTERA_DISP::edid_find
//      Find the EDID block for a monitor
//
//  Description:
//      This function tries to find the EDID block for a monitor from the
//      registry. This function looks at the path pointed to by the specified
//      device key. The EDID block is stored in a registry key.
//
#if 0 // TBD - need to implement
BOOL cTERA_DISP::edid_find(
    WCHAR *mon_device_key,
    BYTE  *edid,
    DWORD  len_edid
    )
{
    BOOL            ret = FALSE;
    SP_DEVINFO_DATA devInfoData;
    DWORD           length = len_edid;
    HDEVINFO        devInfo = NULL;
    BOOL            found = FALSE;

    devInfo = SetupDiGetClassDevsEx(&GUID_CLASS_MONITOR, NULL, NULL,
                                    DIGCF_PRESENT, NULL, NULL, NULL);
    if (devInfo != NULL)
    {
        // Go through all the registry entries in the monitor class
        for (int i = 0; GetLastError() != ERROR_NO_MORE_ITEMS; i++)
        {
            mTERA_MEM_ZERO(&devInfoData, sizeof(devInfoData));
            devInfoData.cbSize = sizeof(devInfoData);

            if (SetupDiEnumDeviceInfo(devInfo, i, &devInfoData))
            {
                WCHAR drv_key[MAX_DEVICE_KEY_LENGTH];

                //
                // Look in the registry for the device's software key (aka
                // driver key)
                //
                if (SetupDiGetDeviceRegistryProperty(devInfo, &devInfoData,
                                                      SPDRP_DRIVER, NULL,
                                                      (PBYTE)(&drv_key),
                                                      MAX_DEVICE_KEY_LENGTH,
                                                      NULL))
                {
                    if (wcscmp(wcsstr(mon_device_key, L"{"), drv_key) == 0)
                    {
                        mTERA_LOG_DBG("Found a match for '%s'", drv_key);
                        found = TRUE;
                        break;
                    }
                }
                else
                {
                    mTERA_LOG_DBG("SetupDiGetDeviceRegistryProperty() returned FALSE");
                }
            }
            else
            {
                mTERA_LOG_DBG("SetupDiEnumDeviceInfo() returned FALSE");
            }
        } // end for
    } // end if
    else
    {
        mTERA_LOG_DBG("SetupDiGetClassDevsEx() returned NULL");
    }

    if (found)
    {
        HKEY hDevRegKey = SetupDiOpenDevRegKey(devInfo, &devInfoData,
                                               DICS_FLAG_GLOBAL, 0, DIREG_DEV,
                                               KEY_QUERY_VALUE);
        if (hDevRegKey != INVALID_HANDLE_VALUE)
        {
            RegQueryValueEx(hDevRegKey, L"EDID", NULL, NULL, edid, &length);
            RegCloseKey(hDevRegKey);
            ret = TRUE;
        }
    }
    else
    {
        mTERA_LOG_DBG("could not find monitor in registry");
    }

    return ret;
} // cTERA_DISP::edid_find()
#endif

// --------------------------------------------------------------------------
//  cTERA_DISP::vid_extract
//      Extract the vendor ID
//
//  Description:
//      This function extracts the VID from a byte sequence (that is part of
//      the EDID block) and writes it into the specified buffer in ASCII format.
//
void cTERA_DISP::vid_extract(
    char *vid,
    BYTE *edid
    )
{
    vid[0] = ((edid[EDID_VID_BYTE1_OFFSET] >> 2) & 0x1F) + 'A' - 1;
    vid[1] = (((edid[EDID_VID_BYTE1_OFFSET] << 3) & 0x18) |
              ((edid[EDID_VID_BYTE2_OFFSET] >> 5) & 0x7)) + 'A' - 1;
    vid[2] = (edid[EDID_VID_BYTE2_OFFSET] & 0x1F) + 'A' - 1;
    vid[3] = 0; // NULL-terminate
} // cTERA_DISP::vid_extract()

// --------------------------------------------------------------------------
//  cTERA_DISP::pid_extract
//      Extract the product ID
//
//  Description:
//      This function extracts the PID from a byte sequence (that is part of
//      the EDID block) and writes it into the specified buffer in ASCII format.
//
void cTERA_DISP::pid_extract(
    char *pid,
    BYTE *edid
    )
{
    // Store the PID as ASCII characters
    sprintf(pid, "%X%02X",
            edid[EDID_PID_BYTE2_OFFSET], edid[EDID_PID_BYTE1_OFFSET]);
} // cTERA_DISP::pid_extract()

// --------------------------------------------------------------------------
//  cTERA_DISP::serial_extract
//      Extract the serial number
//
//  Description:
//      This function extracts the serial number from a byte sequence (that is
//      part of the EDID block) and writes it into the specified buffer in
//      ASCII format. This is not the serial number from a descriptor block.
//
void cTERA_DISP::serial_extract(
    char *serial,
    BYTE *edid
    )
{
    union {
        UINT32 num;
        BYTE   raw[4];
    } extracted;

    extracted.raw[0] = edid[0];
    extracted.raw[1] = edid[1];
    extracted.raw[2] = edid[2];
    extracted.raw[3] = edid[3];

    sprintf(serial, "%d", extracted.num);
} // cTERA_DISP::serial_extract()

// --------------------------------------------------------------------------
//  cTERA_DISP::edid_desc_block_parse
//      Parse an EDID descriptor block
//
//  Description:
//      This function parses an EDID descriptor block for any names or serial
//      numbers. Each descriptor block contains a header to indicate what it
//      holds.
//
//      The buffers for the serial number and monitor name MUST be at least
//      14 bytes in length (13 bytes for the string + 1 byte for the NULL
//      character). This function will always set the last character of the
//      buffers to be NULL.
//
void cTERA_DISP::edid_desc_block_parse(
    char  serial[14],
    char  name[14],
    BYTE *edid
    )
{
    UINT32 desc_block_hdr = 0;
    INT    i;

    mTERA_MEM_COPY(&desc_block_hdr, edid, sizeof(desc_block_hdr));
    if (desc_block_hdr == EDID_DESC_BLOCK_HDR_SERIAL)
    {
        // Go through the max length of the descriptor block (13 characters)
        for (i = 0; i < 13; i++)
        {
            serial[i] = (char)edid[EDID_DESC_BLOCK_DATA_OFFSET + i];
            if (serial[i] == '\n')
            {
                // The line-feed character means end of string in EDID spec
                serial[i] = '\0';
            }
        }

        // Make sure the last character is always NULL
        serial[13] = '\0';
    }
    else if (desc_block_hdr == EDID_DESC_BLOCK_HDR_NAME)
    {
        for (i = 0; i < 13; i++)
        {
            name[i] = (char)edid[EDID_DESC_BLOCK_DATA_OFFSET + i];
            if (name[i] == '\n')
            {
                // The line-feed character means end of string in EDID spec
                name[i] = '\0';
            }
        }

        // Make sure the last character is always NULL
        name[13] = '\0';
    }
} // cTERA_DISP::edid_desc_block_parse()

// --------------------------------------------------------------------------
//  cTERA_DISP::edid_parse
//      Parse the EDID block
//
//  Description:
//      This function parses the EDID block for a monitor. This function first
//      checks to see if the EDID block is valid by looking at the EDID header.
//      If the header is valid, then this function will try to extract the
//      VID, PID, serial number, and name from the block and store it into the
//      monitor entry.
//
#if 0 // TBD - need to implement
BOOL cTERA_DISP::edid_parse(
    WCHAR *mon_device_key,
    sTERA_DISP_EDID_ID *mon_edid_id
    )
{
    BOOL   ret = FALSE;
    UINT64 edid_hdr = 0;
    BYTE   edid[1024];

    // Find the EDID entry in the registry
    ret = edid_find(mon_device_key, edid, sizeof(edid));
    if (ret)
    {
        // Check header to see if its a valid EDID
        mTERA_MEM_COPY(&edid_hdr, &edid[0], sizeof(edid_hdr));
        if (edid_hdr == EDID_HDR)
        {
            // Extract the VID
            vid_extract(mon_edid_id->vid, edid);

            // Extract the PID
            pid_extract(mon_edid_id->pid, edid);

            //
            // Parse the descriptor blocks for a serial number and name if they
            // exist. The strings are initialized to NULL strings to deal with
            // the scenario where the name and/or serial number are not found
            // in the EDID.
            //
            strcpy_s(mon_edid_id->name, "\0");
            strcpy_s(mon_edid_id->serial, "\0");

            edid_desc_block_parse(mon_edid_id->serial, mon_edid_id->name,
                                  &edid[EDID_DESC_BLOCK1_OFFSET]);
            edid_desc_block_parse(mon_edid_id->serial, mon_edid_id->name,
                                  &edid[EDID_DESC_BLOCK2_OFFSET]);
            edid_desc_block_parse(mon_edid_id->serial, mon_edid_id->name,
                                  &edid[EDID_DESC_BLOCK3_OFFSET]);
            edid_desc_block_parse(mon_edid_id->serial, mon_edid_id->name,
                                  &edid[EDID_DESC_BLOCK4_OFFSET]);

            //
            // The descriptor blocks did not contain a serial, try the serial
            // field
            //
            if (mon_edid_id->serial[0] == '\0')
            {
                mTERA_LOG_DBG("looking for serial number in serial number field");

                serial_extract(mon_edid_id->serial, &edid[EDID_SERIAL_OFFSET]);
            }

            mTERA_LOG_DBG("name='%s', serial='%s', vid='%s', pid='%s'",
                           mon_edid_id->name,
                           mon_edid_id->serial,
                           mon_edid_id->vid,
                           mon_edid_id->pid);
        }
        else
        {
            mTERA_LOG_ERR("bad EDID entry: %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X",
                          edid[0], edid[1], edid[2], edid[3],
                          edid[4], edid[5], edid[6], edid[7],
                          edid[8], edid[9], edid[10], edid[11],
                          edid[12], edid[13], edid[14], edid[15]);
            ret = FALSE;
        }
    }
    else
    {
        mTERA_LOG_ERR("could not open registry");
    }

    return ret;
} // cTERA_DISP::edid_parse()
#endif

// --------------------------------------------------------------------------
//  cTERA_DISP::edid_match_found
//      Look through the client topology for a display adapters monitor
//
//  Description:
//      This function is given the EDID identification parameters of the
//      monitor attached to a display adapter. It also receives the client
//      supplied display topology. It searches the client topology for the
//      display adapter monitor. If the monitor is found the function returns
//      a non-zero value along with a value that is the index of the client
//      topology entry for the matched monitor. If the monitor is not found
//      this function returns 0.
//
INT cTERA_DISP::edid_match_found(
    sTERA_DISP_EDID_ID  *sys_disp_edid,
    sTERA_DISP_EDID_ID  *client_disp_edid,
    INT32               *client_topo_idx
    )
{
    INT ret = 0;
    INT i;

    // cancel the search if the system adapter information is not valid
    if ((strlen(sys_disp_edid->name) > 0) ||
        (strlen(sys_disp_edid->serial) > 0) ||
        (strlen(sys_disp_edid->vid) > 0) ||
        (strlen(sys_disp_edid->pid) > 0))
    {
        mTERA_LOG_DBG("sys monitor name='%s', serial='%s', vid='%s', pid='%s'",
                      sys_disp_edid->name,
                      sys_disp_edid->serial,
                      sys_disp_edid->vid,
                      sys_disp_edid->pid);

        // loop through the client topology entries until a match is found
        for (i=0; i<TERA_DISP_MAX_DISP; i++)
        {
            // display information for valid monitors
            if ((strlen(client_disp_edid[i].name) > 0) ||
                (strlen(client_disp_edid[i].vid) > 0) ||
                (strlen(client_disp_edid[i].pid) > 0) ||
                (strlen(client_disp_edid[i].serial) > 0))
            {
                mTERA_LOG_DBG("client monitor name='%s', serial='%s', vid='%s', pid='%s'",
                              client_disp_edid[i].name,
                              client_disp_edid[i].serial,
                              client_disp_edid[i].vid,
                              client_disp_edid[i].pid);
            }

            if (!strncmp(sys_disp_edid->vid, client_disp_edid[i].vid, TERA_MON_EDID_VID_LEN) &&
                !strncmp(sys_disp_edid->pid, client_disp_edid[i].pid, TERA_MON_EDID_PID_LEN) &&
                !strncmp(sys_disp_edid->serial, client_disp_edid[i].serial, TERA_MON_EDID_SERIAL_LEN))
            {
                *client_topo_idx = i;
                mTERA_LOG_DBG("EDID match found in client topo entry %d", *client_topo_idx);
                ret = 1;
                break;
            }
        }
    }

    return ret;
} // cTERA_DISP::edid_match_found()

// --------------------------------------------------------------------------
//  cTERA_DISP::dump_display_mode
//      Dump the contents of the mode structure
//
//  Description:
//      This function dumps the contents of the mode structure.
//
#if 0 // TBD - need to implement
void cTERA_DISP::dump_display_mode(
    DEVMODE *mode
    )
{
    mTERA_LOG_DBG("              %d x %d @ (%d, %d) rot=%d freq=%d",
                  mode->dmPelsWidth,
                  mode->dmPelsHeight,
                  mode->dmPosition.x,
                  mode->dmPosition.y,
                  mode->dmDisplayOrientation,
                  mode->dmDisplayFrequency);
#if 1
    mTERA_LOG_DBG("              dmFields=0x%08X,             dmBitsPerPel=0x%08X",
                  mode->dmFields, mode->dmBitsPerPel);
    mTERA_LOG_DBG("              dmDisplayFixedOutput=0x%08X, dmDisplayFlags=0x%08X",
                  mode->dmDisplayFixedOutput, mode->dmDisplayFlags);
    mTERA_LOG_DBG("              dmDitherType=0x%08X          dmLogPixels=0x%08X",
                  mode->dmDitherType, mode->dmLogPixels);
    mTERA_LOG_DBG("              dmPanningWidth=0x%08X        dmPanningHeight=0x%08X",
                  mode->dmPanningWidth, mode->dmPanningHeight);
    mTERA_LOG_DBG("              dmYResolution=0x%08X,        dmScale=0x%08X",
                  mode->dmYResolution, mode->dmScale);
    mTERA_LOG_DBG("              dmSize=0x%08X", mode->dmSize);
#endif

} // cTERA_DISP::dump_display_mode()
#endif

// --------------------------------------------------------------------------
//  cTERA_DISP::dump_display_device
//      Dump the contents of the DISPLAY_DEVICE structure
//
//  Description:
//      This function dumps the contents of the DISPLAY_DEVICE structure.
//
#if 0 // TBD - need to implement
void cTERA_DISP::dump_display_device(
    DISPLAY_DEVICE *dev
    )
{
    mTERA_LOG_DBG("  Name '%s'", dev->DeviceName);
    mTERA_LOG_DBG("  String '%s'", dev->DeviceString);
    mTERA_LOG_DBG("  Id '%s'", dev->DeviceID);
//    mTERA_LOG_DBG("  Key '%s'", dev->DeviceKey);
    mTERA_LOG_DBG("  StateFlags=0x%08X", dev->StateFlags);
} // cTERA_DISP::dump_display_device()
#endif

// --------------------------------------------------------------------------
//  cTERA_DISP::is_client_mapping_valid
//      Validates the array that maps display adapters to client displays
//
//  Description:
//      This function validates the array that maps display adapters to client
//      displays. It looks through all of the entries and makes sure that none
//      of them are identical. The function returns a non-zero value if the
//      array is valid otherwise it returns 0.
//
#if 0 // TBD - need to implement
INT cTERA_DISP::is_client_mapping_valid(
    INT32   *map,
    DEVMODE *mode
    )
{
    INT     i;
    INT     j;

    for (i=0; i<TERA_DISP_MAX_DISP; i++)
    {
        for (j=i+1; j<TERA_DISP_MAX_DISP; j++)
        {
            if ((map[i] != NO_MAPPING_FOUND) &&
                (map[i] != ADAPTER_WITHOUT_MONITOR) &&
                (map[j] == map[i]))
            {
                mTERA_LOG_DBG("duplicate mapping found for adapter[%d]=%d and adapter[%d]=%d",
                              i, map[i], j, map[j]);

                // When duplicate entries are found check if only one display
                // is ON. If so, clear the mapping for the display that is OFF.
                if ( (mode[i].dmFields & DM_POSITION) &&
                    !(mode[j].dmFields & DM_POSITION))
                {
                    mTERA_LOG_DBG("cleared mapping for adapter[%d] because it was off", j);
                    map[j] = NO_MAPPING_FOUND;
                }
                else if (!(mode[i].dmFields & DM_POSITION) &&
                          (mode[j].dmFields & DM_POSITION))
                {
                    mTERA_LOG_DBG("cleared mapping for adapter[%d] because it was off", i);
                    map[i] = NO_MAPPING_FOUND;
                }
                else
                {
                    mTERA_LOG_DBG("dmFields for adapter[%d]=0x%08X and adapter[%d]=0x%08X",
                                  i,
                                  mode[i].dmFields,
                                  j,
                                  mode[j].dmFields);
                    return 0;
                }
            }
        }
    }
    return 1;
} // cTERA_DISP::is_client_mapping_valid()
#endif

// --------------------------------------------------------------------------
//  cTERA_DISP::is_monitor_attached
//      Returns indication of whether a monitor is attached
//
//  Description:
//      This function returns 1 if a monitor is attached and 0 if a monitor is
//      not attached. The function considers a monitor to be attached if the
//      DISPLAY_DEVICE_ATTACHED and DISPLAY_DEVICE_ACTIVE are set in the
//      StateFlags structure AND the length of the DeviceID and DeviceString
//      are greater than 1 AND the characters '{' and '}' are included in the
//      DeviceID.
//
//      The string length check is included because sometimes when a session
//      first starts the StateFlags are set but the strings are empty. I've
//      also seen a problem where the DeviceID string was partially complete.
//      The Device ID should include the device GUID. To ensure this is the
//      case the code verifies the '{' and '}' characters are included in
//      the string. If the checks on the strings are not included the host
//      SW will encounter an error where the agent hangs or crashes when the
//      mon structure is passed to the function SetupDiGetClassDevsEx().
//
#if 0 // TBD - need to implement
INT cTERA_DISP::is_monitor_attached(
    DISPLAY_DEVICE  *mon
    )
{
    if ((mon->StateFlags & DISPLAY_DEVICE_ATTACHED) &&
        (wcslen(mon->DeviceID) > 1) &&
        (wcslen(mon->DeviceString) > 1) &&
        (wcschr(mon->DeviceID, L'{') != NULL) &&
        (wcschr(mon->DeviceID, L'}') != NULL))
    {
        return 1;
    }
    else
    {
        return 0;
    }
} // cTERA_DISP::is_monitor_attached()
#endif

// --------------------------------------------------------------------------
//  cTERA_DISP::is_adapter_active_no_monitor
//      Returns indication of whether an adapter is active but nothing is attached
//
//  Description:
//      This function returns 1 if a display adpater is active but no monitor is
//      attached.
//
//      The string length check is included because sometimes when a session
//      first starts the StateFlags are set but the strings are empty. I've
//      also seen a problem where the DeviceID string was partially complete.
//      The Device ID should include the device GUID. To ensure this is the
//      case the code verifies the '{' and '}' characters are included in
//      the string. If the checks on the strings are not included the host
//      SW will encounter an error where the agent hangs or crashes when the
//      mon structure is passed to the function SetupDiGetClassDevsEx().
//
#if 0 // TBD - need to implement
INT cTERA_DISP::is_adapter_active_no_monitor(
    DISPLAY_DEVICE  *mon
    )
{
    if (!(mon->StateFlags & DISPLAY_DEVICE_ATTACHED) &&
        (mon->StateFlags & DISPLAY_DEVICE_ACTIVE) &&
        (wcslen(mon->DeviceID) > 1) &&
        (wcslen(mon->DeviceString) > 1) &&
        (wcschr(mon->DeviceID, L'{') != NULL) &&
        (wcschr(mon->DeviceID, L'}') != NULL))
    {
        return 1;
    }
    else
    {
        return 0;
    }
} // cTERA_DISP::is_adapter_active_no_monitor()
#endif
