////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     PCoIP Host Software Network Property Page Class Implementation
///
///     This file implements the PCoIP Host Software Network property page
///     class.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include <wx/wxprec.h>

#if !defined(WX_PRECOMP)
    #include <wx/wx.h>
#endif

#include "tera_errors.h"
#include "tera_log.h"
#include "tera_types.h"

#include "tera_pcoip_agent_internal.h"
#include "tera_pcoip_agent_gui_prop_net.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

BEGIN_EVENT_TABLE(cPCOIP_AGENT_GUI_PROP_NET, wxPanel)
    EVT_BUTTON(wxID_REFRESH_BTN, cPCOIP_AGENT_GUI_PROP_NET::evt_refresh_btn)
    EVT_INIT_DIALOG(cPCOIP_AGENT_GUI_PROP_NET::evt_init_dialog)
END_EVENT_TABLE()

// ---- Local Variables -----------------------------------------------------

// ---- Local Function Prototypes -------------------------------------------

void cPCOIP_AGENT_GUI_PROP_NET::evt_init_dialog(
    wxInitDialogEvent& evt
    )
{
    mTERA_LOG_DBG("NET EVT_INIT_DIALOG");

    // Populate the property grid with the current values
    prop_grid_populate();

    evt.Skip();
} // cPCOIP_AGENT_GUI_PROP_NET::evt_init_dialog()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_NET::evt_refresh_btn
//      Process the refresh button click event
//
//  Description:
//      This function processes the refresh button click event. This function
//      calls updates the property grid with the most up-to-date network
//      configuration.
//
void cPCOIP_AGENT_GUI_PROP_NET::evt_refresh_btn(
    wxCommandEvent& evt
    )
{
    // Populate the property grid with the current values
    prop_grid_populate();
} // cPCOIP_AGENT_GUI_PROP_NET::evt_refresh_btn()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_SESS::prop_grid_populate
//      Populate the property grid with the latest network information
//
//  Description:
//      This function populates the property grid with the latest network
//      information. If any of the fields in the property grid are not
//      available, then the value is set to "Unavailable".
//
void cPCOIP_AGENT_GUI_PROP_NET::prop_grid_populate(void)
{
    sTERA_PCOIP_HCC_APDU_HMI_VAR_IPV4_CFG  ipv4_cfg;
    sTERA_PCOIP_HCC_APDU_HMI_VAR_LINK_INFO link_info;
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK sess_link;
    wxString                               val;
    int                                    display_help = 0;

    mTERA_LOG_DBG("start");

    // Get the network link information
    if (wxGetApp().gui().pcoip_link().link_info_get(&link_info) == TERA_SUCCESS)
    {
        if ((link_info.state == TERA_PCOIP_HCC_APDU_HMI_LINK_STATE_UP) &&
            (link_info.rate != TERA_PCOIP_HCC_APDU_HMI_LINK_RATE_NONE))
        {
            val.Printf(_T("%S Mbps, %S Duplex"),
                       link_info.rate == TERA_PCOIP_HCC_APDU_HMI_LINK_RATE_10M ? "10" :
                       link_info.rate == TERA_PCOIP_HCC_APDU_HMI_LINK_RATE_100M ? "100" :
                       "1000",
                       (link_info.flags & TERA_PCOIP_HCC_APDU_HMI_VAR_LINK_INFO_FLAGS_FULL_DUPLEX) ? "Full" : "Half");

            m_prop_grid->SetPropertyValue(_T("gen_link"), val);
        }
        else
        {
            m_prop_grid->SetPropertyValue(_T("gen_link"), _T("Disconnected"));
        }
        val.Printf(_T("%S"), link_info.mac);
        m_prop_grid->SetPropertyValue(_T("gen_mac"), val);
    }
    else
    {
        m_prop_grid->SetPropertyValue(_T("gen_link"), _("Unavailable"));

        // Display the help tooltips
        display_help = 1;
    }

    // Get the IPv4 configuration
    if (wxGetApp().gui().pcoip_link().ipv4_cfg_get(&ipv4_cfg) ==
        TERA_SUCCESS)
    {
        m_prop_grid->SetPropertyValue(_T("gen_prot"),
                                      (ipv4_cfg.flags & TERA_PCOIP_HCC_APDU_HMI_VAR_IPV4_CFG_FLAGS_IPV4_EN) ? _("IPv4") :_("IPv6"));
        m_prop_grid->SetPropertyValue(_T("ipv4_dhcp_en"),
                                      (ipv4_cfg.flags & TERA_PCOIP_HCC_APDU_HMI_VAR_IPV4_CFG_FLAGS_DHCP_EN) ? _("True") :_("False"));
        val.Printf(_T("%S"), ipv4_cfg.ip);
        m_prop_grid->SetPropertyValue(_T("ipv4_ip"), val);
        val.Printf(_T("%S"), ipv4_cfg.mask);
        m_prop_grid->SetPropertyValue(_T("ipv4_mask"), val);
        val.Printf(_T("%S"), ipv4_cfg.gw);
        m_prop_grid->SetPropertyValue(_T("ipv4_gw"), val);
        val.Printf(_T("%S"), ipv4_cfg.pri_dns);
        m_prop_grid->SetPropertyValue(_T("ipv4_pri_dns"), val);
        val.Printf(_T("%S"), ipv4_cfg.sec_dns);
        m_prop_grid->SetPropertyValue(_T("ipv4_sec_dns"), val);
    }
    else
    {
        m_prop_grid->SetPropertyValue(_T("gen_mac"), _("Unavailable"));
        m_prop_grid->SetPropertyValue(_T("gen_prot"), _("Unavailable"));
        m_prop_grid->SetPropertyValue(_T("ipv4_dhcp_en"), _("Unavailable"));
        m_prop_grid->SetPropertyValue(_T("ipv4_ip"), _("Unavailable"));
        m_prop_grid->SetPropertyValue(_T("ipv4_mask"), _("Unavailable"));
        m_prop_grid->SetPropertyValue(_T("ipv4_gw"), _("Unavailable"));
        m_prop_grid->SetPropertyValue(_T("ipv4_pri_dns"), _("Unavailable"));
        m_prop_grid->SetPropertyValue(_T("ipv4_sec_dns"), _("Unavailable"));

        // Display the help tooltips
        display_help = 1;
    }

    // Get the network link information
    if (wxGetApp().gui().pcoip_link().sess_link_get(&sess_link) == TERA_SUCCESS)
    {
        if (sess_link.flags & TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK_FLAGS_CONNECTED)
        {
            val.Printf(_T("%S"), sess_link.ip);
            m_prop_grid->SetPropertyValue(_T("peer_ip"), val);

            val.Printf(_T("%S"),sess_link.mac);
            m_prop_grid->SetPropertyValue(_T("peer_mac"), val);
        }
        else
        {
            m_prop_grid->SetPropertyValue(_T("peer_ip"), _("Disconnected"));
        }
    }
    else
    {
        m_prop_grid->SetPropertyValue(_T("peer_ip"), _("Unavailable"));
        m_prop_grid->SetPropertyValue(_T("peer_mac"), _("Unavailable"));

        // Display the help tooltips
        display_help = 1;
    }

    // Setup and show the help tooltips
    help_btn_setup(display_help);

    mTERA_LOG_DBG("end");
} // cPCOIP_AGENT_GUI_PROP_NET::prop_grid_populate()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_NET::help_btn_setup
//      Set up and show the tooltip help button
//
//  Description:
//      This function sets up and shows the tooltip help button. The button
//      is formatted to look like plain text. When the user hovers or clicks
//      the text, a tooltip will appear. The tooltip provides possible
//      causes and solutions to why certain features are unavailable.
//
void cPCOIP_AGENT_GUI_PROP_NET::help_btn_setup(int display_help)
{
    if (!display_help)
    {
        // Hide the help text
        m_help_link->Show(false);
    }
    else
    {
        // Show the help text and set it to something generic first
        m_help_link->SetToolTip(IDS_STR_HELP_GEN_NO_SUPPORT);

        // ...then determine the best help text if it exists
        if (!wxGetApp().gui().pcoip_link().is_ready())
        {
            m_help_link->SetToolTip(IDS_STR_HELP_GEN_NO_LINK);
        }
        m_help_link->Show(true);
    }
} // cPCOIP_AGENT_GUI_PROP_NET::help_btn_setup()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_NET::cPCOIP_AGENT_GUI_PROP_NET
///     Constructor
//
//  Description:
///     This is the PCoIP Host Software Network property page class
///     constructor.
///
/// \return
///     None
///
cPCOIP_AGENT_GUI_PROP_NET::cPCOIP_AGENT_GUI_PROP_NET(
    wxWindow *parent
    ) :
    cPCOIP_AGENT_GUI_PROP_NET_BASE(parent)
{
    mTERA_LOG_DBG("start");

    m_prop_grid->Append(new wxPropertyCategory(_("General")));

    m_prop_grid->Append(new wxStringProperty(_("MAC Address"),
                                             _T("gen_mac")));
    m_prop_grid->Append(new wxStringProperty(_("Network Link"),
                                             _T("gen_link")));
    m_prop_grid->Append(new wxStringProperty(_("IP Protocol"),
                                             _T("gen_prot")));

    m_prop_grid->Append(new wxPropertyCategory(_("IPv4")));

    m_prop_grid->Append(new wxStringProperty(_("Enable DHCP"),
                                             _T("ipv4_dhcp_en")));
    m_prop_grid->Append(new wxStringProperty(_("IP Address"),
                                             _T("ipv4_ip")));
    m_prop_grid->Append(new wxStringProperty(_("Subnet Mask"),
                                             _T("ipv4_mask")));
    m_prop_grid->Append(new wxStringProperty(_("Default Gateway"),
                                             _T("ipv4_gw")));
    m_prop_grid->Append(new wxStringProperty(_("Primary DNS Server"),
                                             _T("ipv4_pri_dns")));
    m_prop_grid->Append(new wxStringProperty(_("Secondary DNS Server"),
                                             _T("ipv4_sec_dns")));

    m_prop_grid->Append(new wxPropertyCategory(_("Peer")));

    m_prop_grid->Append(new wxStringProperty(_("IP Address"),
                                             _T("peer_ip")));
    m_prop_grid->Append(new wxStringProperty(_("MAC Address"),
                                             _T("peer_mac")));

    // Fit first column to width of longest label
    m_prop_grid->FitColumns();

    mTERA_LOG_DBG("end");
} // cPCOIP_AGENT_GUI_PROP_NET::cPCOIP_AGENT_GUI_PROP_NET()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_NET::~cPCOIP_AGENT_GUI_PROP_NET
///     Destructor
//
//  Description:
///     This is the PCoIP Host Software Network property page class destructor.
///     This function cleans up any loose ends before being destroyed.
///
/// \return
///     None
///
cPCOIP_AGENT_GUI_PROP_NET::~cPCOIP_AGENT_GUI_PROP_NET(void)
{
    mTERA_LOG_DBG("start");

    mTERA_LOG_DBG("end");
} // cPCOIP_AGENT_GUI_PROP_NET::~cPCOIP_AGENT_GUI_PROP_NET()

