////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     Link Class Implementation
///
///     This file implements the link class, which provides userspace
///     applications a way to talk to blade and client devices via device
///     drivers in kernel space. This class automatically establishes a link
///     to the device driver interface. Then through ioctl calls, this class
///     gathers all the supported features available to the host OS from the
///     blade and client devices.
///
///     The available features from the blade/client are:
///         - environment
///         - firmware
///         - local cursor
///         - local keyboard
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include "tera_types.h"
#include "tera_errors.h"
#include "tera_log.h"
#include "tera_mem.h"
#include "tera_pcoip_link.h"
#include "tera_pcoip_hcc_ioctl.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

#define PCOIP_HCC_DEV_PATH  "/dev/pcoip_host0"
#define PCOIP_HDC_DEV_PATH  "/dev/pcoip_host1"

// ---- Local Variables -----------------------------------------------------

// ---- Protected Functions -------------------------------------------------

// ---- Public Functions ----------------------------------------------------

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::cTERA_PCOIP_LINK
///     Constructor
//
//  Description:
///     This function is the class constructor. All member data variables are
///     initialized and a feature set update is performed.
///
cTERA_PCOIP_LINK::cTERA_PCOIP_LINK(void)
{
    INT32 ret;

    mTERA_LOG_DBG("start");

    // Open the PCoIP HCC device
    ret = m_hcc_dev.init(PCOIP_HCC_DEV_PATH);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("failed to initialize PCoIP HCC device object (ret=%d)",
                      ret);
    }
    else
    {
        mTERA_LOG_DBG("successfully opened PCoIP HCC object");
    }

    // Open the PCoIP HDC device
    ret = m_hdc_dev.init(PCOIP_HDC_DEV_PATH);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("failed to initialize PCoIP HDC device object (ret=%d)",
                      ret);
    }
    else
    {
        mTERA_LOG_DBG("successfully opened PCoIP HDC object");
    }

    // Initialize all values
    m_started = 0;
    m_kmp_pri_ctxt_valid = 0;


    mTERA_MEM_ZERO(&m_kmp_pri_ctxt, sizeof(m_kmp_pri_ctxt));
    mTERA_MEM_ZERO(&m_kmp_bmp_cache, sizeof(m_kmp_bmp_cache));

    // Reset the list of available HMI commands and variables
    hmi_avail_cmd_reset();

    // Get supported HMI commands and variables
    hmi_avail_cmd_get();

    mTERA_LOG_DBG("end");
} // cTERA_PCOIP_LINK::cTERA_PCOIP_LINK()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::~cTERA_PCOIP_LINK
///     Destructor
//
//  Description:
///     This function performs clean up of this class, such as freeing the
///     memory used for the host monitor information.
///
/// \return
///     None
///
cTERA_PCOIP_LINK::~cTERA_PCOIP_LINK(void)
{
    mTERA_LOG_DBG("start");

//    unhide_cursor();

    mTERA_LOG_DBG("end");
} // cTERA_PCOIP_LINK::~cTERA_PCOIP_LINK()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::is_ready
///     Return if this object is ready to talk to the devices
//
//  Description:
///     This function returns if this object is ready to talk to the devices.
///     If either device is not ready, then this function returns 0.
///
INT cTERA_PCOIP_LINK::is_ready(void)
{
    return (m_hcc_dev.is_ready() && m_hdc_dev.is_ready());
} // cTERA_PCOIP_LINK::is_ready()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::link_heartbeat
///     Provide a heartbeat to this object
//
//  Description:
///     This function is for the caller to provide a heartbeat to this object.
///     This function sets sess_link_change to a non-zero value if the session
///     link state changes. It also updates sess_link_flags when the session
///     link state changes.
///
void cTERA_PCOIP_LINK::link_heartbeat(
    INT                                    *sess_link_change,
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK *sess_link
    )
{
    INT32                                  ret;
    static INT                             first_time = 1;
    static UINT8                           transition_cnt = 0;

    *sess_link_change = 0;

    // Check for session transitions
    ret = sess_link_get(sess_link);
    if (ret == TERA_SUCCESS)
    {
        if ((sess_link->transition_cnt != transition_cnt) || first_time)
        {
            first_time = 0;
            transition_cnt = sess_link->transition_cnt;

            // Reset KMP PRI context
            kmp_pri_ctxt_reset();

            if (sess_link->flags &
                TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK_FLAGS_CONNECTED)
            {
                // Get the KMP PRI context and init the cached bitmap storage
                kmp_pri_ctxt_get();
                init_bitmap_cache();
            }

            // Notify users of a session link change
            *sess_link_change = 1;
            mTERA_LOG_DBG("sess link state changed - flags: 0x%08X, trans_cnt=%d, IP=%s",
                          sess_link->flags,
                          sess_link->transition_cnt,
                          sess_link->ip);
        }
    }
    else
    {
        // Reset KMP PRI context
        kmp_pri_ctxt_reset();
    }
} // cTERA_PCOIP_LINK::link_heartbeat()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::start
///     Start the communications with the PCoIP device
//
//  Description:
///     This function starts the communications with the PCoIP device. This
///     function MUST be called before calling
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::start(void)
{
    INT32 ret = TERA_SUCCESS;

    if (m_started)
    {
        mTERA_LOG_ERR("already started");
    }
    else
    {
        // Finished starting up
        m_started = 1;
    }

    return ret;
} // cTERA_PCOIP_LINK::start()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::stop
///     Stop the notifications
//
//  Description:
///     This function stops the notifications that this object gets from the
///     device. This function must be called prior to destroying this object.
///     This function cannot be called in a destructor because it potentially
///     blocks while other threads terminate.
///
void cTERA_PCOIP_LINK::stop(void)
{
    mTERA_LOG_DBG("start");

    if (m_started)
    {
        // close the HDC channel
        hdc_close();

        // Stopped
        m_started = 0;
    }

    mTERA_LOG_DBG("end");
} // cTERA_PCOIP_LINK::stop()
